import warnings
from copy import copy

import numpy as np

from hdmf.common import DynamicTable, AlignedDynamicTable
from hdmf.utils import docval, popargs, popargs_to_dict, get_docval, getargs

from . import register_class, CORE_NAMESPACE
from .base import TimeSeries, TimeSeriesReferenceVectorData
from .core import NWBContainer
from .device import Device


def ensure_unit(self, name, current_unit, unit, nwb_version):
    """A helper to ensure correct unit used.

    Issues a warning with details if `current_unit` is to be ignored, and
    `unit` to be used instead.
    """
    if current_unit != unit:
        warnings.warn(
            "Unit '%s' for %s '%s' is ignored and will be set to '%s' "
            "as per NWB %s."
            % (current_unit, self.__class__.__name__, name, unit, nwb_version))
    return unit


@register_class('IntracellularElectrode', CORE_NAMESPACE)
class IntracellularElectrode(NWBContainer):
    """Describes an intracellular electrode and associated metadata."""

    __nwbfields__ = ('cell_id',
                     'slice',
                     'seal',
                     'description',
                     'location',
                     'resistance',
                     'filtering',
                     'initial_access_resistance',
                     'device')

    @docval({'name': 'name', 'type': str, 'doc': 'the name of this electrode'},
            {'name': 'device', 'type': Device, 'doc': 'the device that was used to record from this electrode'},
            {'name': 'description', 'type': str,
             'doc': 'Recording description, description of electrode (e.g.,  whole-cell, sharp, etc).'},
            {'name': 'slice', 'type': str, 'doc': 'Information about slice used for recording.', 'default': None},
            {'name': 'seal', 'type': str, 'doc': 'Information about seal used for recording.', 'default': None},
            {'name': 'location', 'type': str,
             'doc': 'Area, layer, comments on estimation, stereotaxis coordinates (if in vivo, etc).', 'default': None},
            {'name': 'resistance', 'type': str, 'doc': 'Electrode resistance, unit - Ohm.', 'default': None},
            {'name': 'filtering', 'type': str, 'doc': 'Electrode specific filtering.', 'default': None},
            {'name': 'initial_access_resistance', 'type': str, 'doc': 'Initial access resistance.', 'default': None},
            {'name': 'cell_id', 'type': str, 'doc': 'Unique ID of cell.', 'default': None}
            )
    def __init__(self, **kwargs):
        keys_to_set = (
            'slice',
            'seal',
            'description',
            'location',
            'resistance',
            'filtering',
            'initial_access_resistance',
            'device',
            'cell_id'
        )
        args_to_set = popargs_to_dict(keys_to_set, kwargs)
        super().__init__(**kwargs)

        for key, val in args_to_set.items():
            setattr(self, key, val)


@register_class('PatchClampSeries', CORE_NAMESPACE)
class PatchClampSeries(TimeSeries):
    '''
    Stores stimulus or response current or voltage. Superclass definition for patch-clamp data
    (this class should not be instantiated directly).
    '''

    __nwbfields__ = (
        'electrode',
        'gain',
        'stimulus_description',
        'sweep_number',
    )

    @docval(
        *get_docval(TimeSeries.__init__, 'name'),  # required
        {
            "name": "data",
            "type": ("array_data", "data", TimeSeries),
            "doc": "The data values. The first dimension must be time.",
            "shape": (None,),
        },   # required
        {
            'name': 'unit',
            'type': str,
            'doc': 'The base unit of measurement (should be SI unit)',
        },  # required
        {
            'name': 'electrode',
            'type': IntracellularElectrode,
            'doc': 'IntracellularElectrode group that describes the electrode that was '
                   'used to apply or record this data.',
        },    # required
        {
            'name': 'gain',
            'type': float,
            'doc': 'Units: Volt/Amp (v-clamp) or Volt/Volt (c-clamp)',
        },  # required
        {
            'name': 'stimulus_description',
            'type': str,
            'doc': 'the stimulus name/protocol',
            'default': "N/A",
        },
        *get_docval(
            TimeSeries.__init__,
            'resolution',
            'conversion',
            'timestamps',
            'starting_time',
            'rate',
            'comments',
            'description',
            'control',
            'control_description',
            'offset',
        ),
        {
            'name': 'sweep_number', 'type': (int, 'uint32', 'uint64'),
            'doc': 'Sweep number, allows for grouping different PatchClampSeries '
                   'together via the sweep_table',
            'default': None,
        }
    )
    def __init__(self, **kwargs):
        name, data, unit, stimulus_description = popargs('name', 'data', 'unit', 'stimulus_description', kwargs)
        electrode, gain, sweep_number = popargs('electrode', 'gain', 'sweep_number', kwargs)
        super().__init__(name, data, unit, **kwargs)
        self.electrode = electrode
        self.gain = gain
        self.stimulus_description = stimulus_description

        if sweep_number is not None:
            if not (sweep_number >= 0):
                raise ValueError("sweep_number must be a non-negative integer")

            self.sweep_number = sweep_number


@register_class('CurrentClampSeries', CORE_NAMESPACE)
class CurrentClampSeries(PatchClampSeries):
    '''
    Stores voltage data recorded from intracellular current-clamp recordings. A corresponding
    CurrentClampStimulusSeries (stored separately as a stimulus) is used to store the current
    injected.
    '''

    __nwbfields__ = ('bias_current',
                     'bridge_balance',
                     'capacitance_compensation')

    @docval(*get_docval(PatchClampSeries.__init__, 'name', 'data', 'electrode'),  # required
            {'name': 'gain', 'type': float, 'doc': 'Units - Volt/Volt'},
            *get_docval(PatchClampSeries.__init__, 'stimulus_description'),
            {'name': 'bias_current', 'type': float, 'doc': 'Unit - Amp', 'default': None},
            {'name': 'bridge_balance', 'type': float, 'doc': 'Unit - Ohm', 'default': None},
            {'name': 'capacitance_compensation', 'type': float, 'doc': 'Unit - Farad', 'default': None},
            *get_docval(PatchClampSeries.__init__, 'resolution', 'conversion', 'timestamps', 'starting_time', 'rate',
                        'comments', 'description', 'control', 'control_description', 'sweep_number', 'offset'),
            {'name': 'unit', 'type': str, 'doc': "The base unit of measurement (must be 'volts')",
             'default': 'volts'})
    def __init__(self, **kwargs):
        name, data, unit, electrode, gain = popargs('name', 'data', 'unit', 'electrode', 'gain', kwargs)
        unit = ensure_unit(self, name, unit, 'volts', '2.1.0')
        bias_current, bridge_balance, capacitance_compensation = popargs(
            'bias_current', 'bridge_balance', 'capacitance_compensation', kwargs)
        super().__init__(name, data, unit, electrode, gain, **kwargs)
        self.bias_current = bias_current
        self.bridge_balance = bridge_balance
        self.capacitance_compensation = capacitance_compensation


@register_class('IZeroClampSeries', CORE_NAMESPACE)
class IZeroClampSeries(CurrentClampSeries):
    '''
    Stores recorded voltage data from intracellular recordings when all current and amplifier settings
    are off (i.e., CurrentClampSeries fields will be zero). There is no CurrentClampStimulusSeries
    associated with an IZero series because the amplifier is disconnected and no stimulus can reach
    the cell.
    '''

    __nwbfields__ = ()

    @docval(*get_docval(CurrentClampSeries.__init__, 'name', 'data', 'electrode'),  # required
            {'name': 'gain', 'type': float, 'doc': 'Units: Volt/Volt'},  # required
            {'name': 'stimulus_description', 'type': str,
             'doc': ('The stimulus name/protocol. Setting this to a value other than "N/A" is deprecated as of '
                     'NWB 2.3.0.'),
             'default': 'N/A'},
            *get_docval(CurrentClampSeries.__init__, 'resolution', 'conversion', 'timestamps',
                        'starting_time', 'rate', 'comments', 'description', 'control', 'control_description',
                        'sweep_number', 'offset'),
            {'name': 'unit', 'type': str, 'doc': "The base unit of measurement (must be 'volts')",
             'default': 'volts'})
    def __init__(self, **kwargs):
        name, data, electrode, gain = popargs('name', 'data', 'electrode', 'gain', kwargs)
        bias_current, bridge_balance, capacitance_compensation = (0.0, 0.0, 0.0)
        stimulus_description = popargs('stimulus_description', kwargs)
        stimulus_description = self._ensure_stimulus_description(name, stimulus_description, 'N/A', '2.3.0')
        kwargs['stimulus_description'] = stimulus_description
        super().__init__(name, data, electrode, gain, bias_current, bridge_balance, capacitance_compensation,
                         **kwargs)

    def _ensure_stimulus_description(self, name, current_stim_desc, stim_desc, nwb_version):
        """A helper to ensure correct stimulus_description used.

        Issues a warning with details if `current_stim_desc` is to be ignored, and
        `stim_desc` to be used instead.
        """
        if current_stim_desc != stim_desc:
            warnings.warn(
                "Stimulus description '%s' for %s '%s' is ignored and will be set to '%s' "
                "as per NWB %s."
                % (current_stim_desc, self.__class__.__name__, name, stim_desc, nwb_version))
        return stim_desc


@register_class('CurrentClampStimulusSeries', CORE_NAMESPACE)
class CurrentClampStimulusSeries(PatchClampSeries):
    '''
    Alias to standard PatchClampSeries. Its functionality is to better tag PatchClampSeries for
    machine (and human) readability of the file.
    '''

    __nwbfields__ = ()

    @docval(*get_docval(PatchClampSeries.__init__, 'name', 'data', 'electrode', 'gain'),  # required
            *get_docval(PatchClampSeries.__init__, 'stimulus_description', 'resolution', 'conversion', 'timestamps',
                        'starting_time', 'rate', 'comments', 'description', 'control', 'control_description',
                        'sweep_number', 'offset'),
            {'name': 'unit', 'type': str, 'doc': "The base unit of measurement (must be 'amperes')",
             'default': 'amperes'})
    def __init__(self, **kwargs):
        name, data, unit, electrode, gain = popargs('name', 'data', 'unit', 'electrode', 'gain', kwargs)
        unit = ensure_unit(self, name, unit, 'amperes', '2.1.0')
        super().__init__(name, data, unit, electrode, gain, **kwargs)


@register_class('VoltageClampSeries', CORE_NAMESPACE)
class VoltageClampSeries(PatchClampSeries):
    '''
    Stores current data recorded from intracellular voltage-clamp recordings. A corresponding
    VoltageClampStimulusSeries (stored separately as a stimulus) is used to store the voltage
    injected.
    '''

    __nwbfields__ = ('capacitance_fast',
                     'capacitance_slow',
                     'resistance_comp_bandwidth',
                     'resistance_comp_correction',
                     'resistance_comp_prediction',
                     'whole_cell_capacitance_comp',
                     'whole_cell_series_resistance_comp')

    @docval(*get_docval(PatchClampSeries.__init__, 'name', 'data', 'electrode'),  # required
            {'name': 'gain', 'type': float, 'doc': 'Units - Volt/Amp'},  # required
            *get_docval(PatchClampSeries.__init__, 'stimulus_description'),
            {'name': 'capacitance_fast', 'type': float, 'doc': 'Unit - Farad', 'default': None},
            {'name': 'capacitance_slow', 'type': float, 'doc': 'Unit - Farad', 'default': None},
            {'name': 'resistance_comp_bandwidth', 'type': float, 'doc': 'Unit - Hz', 'default': None},
            {'name': 'resistance_comp_correction', 'type': float, 'doc': 'Unit - percent', 'default': None},
            {'name': 'resistance_comp_prediction', 'type': float, 'doc': 'Unit - percent', 'default': None},
            {'name': 'whole_cell_capacitance_comp', 'type': float, 'doc': 'Unit - Farad', 'default': None},
            {'name': 'whole_cell_series_resistance_comp', 'type': float, 'doc': 'Unit - Ohm', 'default': None},
            *get_docval(PatchClampSeries.__init__, 'resolution', 'conversion', 'timestamps', 'starting_time', 'rate',
                        'comments', 'description', 'control', 'control_description', 'sweep_number', 'offset'),
            {'name': 'unit', 'type': str, 'doc': "The base unit of measurement (must be 'amperes')",
             'default': 'amperes'})
    def __init__(self, **kwargs):
        name, data, unit, electrode, gain = popargs('name', 'data', 'unit', 'electrode', 'gain', kwargs)
        unit = ensure_unit(self, name, unit, 'amperes', '2.1.0')
        capacitance_fast, capacitance_slow, resistance_comp_bandwidth, resistance_comp_correction, \
            resistance_comp_prediction, whole_cell_capacitance_comp, whole_cell_series_resistance_comp = popargs(
                'capacitance_fast', 'capacitance_slow', 'resistance_comp_bandwidth',
                'resistance_comp_correction', 'resistance_comp_prediction', 'whole_cell_capacitance_comp',
                'whole_cell_series_resistance_comp', kwargs)
        super().__init__(name, data, unit, electrode, gain, **kwargs)
        self.capacitance_fast = capacitance_fast
        self.capacitance_slow = capacitance_slow
        self.resistance_comp_bandwidth = resistance_comp_bandwidth
        self.resistance_comp_correction = resistance_comp_correction
        self.resistance_comp_prediction = resistance_comp_prediction
        self.whole_cell_capacitance_comp = whole_cell_capacitance_comp
        self.whole_cell_series_resistance_comp = whole_cell_series_resistance_comp


@register_class('VoltageClampStimulusSeries', CORE_NAMESPACE)
class VoltageClampStimulusSeries(PatchClampSeries):
    '''
    Alias to standard PatchClampSeries. Its functionality is to better tag PatchClampSeries for
    machine (and human) readability of the file.
    '''

    __nwbfields__ = ()

    @docval(*get_docval(PatchClampSeries.__init__, 'name', 'data', 'electrode', 'gain'),  # required
            *get_docval(PatchClampSeries.__init__, 'stimulus_description', 'resolution', 'conversion', 'timestamps',
                        'starting_time', 'rate', 'comments', 'description', 'control', 'control_description',
                        'sweep_number', 'offset'),
            {'name': 'unit', 'type': str, 'doc': "The base unit of measurement (must be 'volts')",
             'default': 'volts'})
    def __init__(self, **kwargs):
        name, data, unit, electrode, gain = popargs('name', 'data', 'unit', 'electrode', 'gain', kwargs)
        unit = ensure_unit(self, name, unit, 'volts', '2.1.0')
        super().__init__(name, data, unit, electrode, gain, **kwargs)


@register_class('SweepTable', CORE_NAMESPACE)
class SweepTable(DynamicTable):
    """
    A SweepTable allows to group PatchClampSeries together which stem from the same sweep.
    A sweep is a group of PatchClampSeries which have the same starting point in time.
    """

    __columns__ = (
            {'name': 'series', 'description': 'PatchClampSeries with the same sweep number',
             'required': True, 'index': True},
            {'name': 'sweep_number', 'description': 'Sweep number of the entries in that row', 'required': True}
    )

    @docval({'name': 'name', 'type': str, 'doc': 'name of this SweepTable', 'default': 'sweep_table'},
            {'name': 'description', 'type': str, 'doc': 'Description of this SweepTable',
             'default': "A sweep table groups different PatchClampSeries together."},
            *get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        warnings.warn("Use of SweepTable is deprecated. Use the IntracellularRecordingsTable "
                      "instead. See also the  NWBFile.add_intracellular_recordings function.",
                      DeprecationWarning)
        super().__init__(**kwargs)

    @docval({'name': 'pcs', 'type': PatchClampSeries,
             'doc': 'PatchClampSeries to add to the table must have a valid sweep_number'})
    def add_entry(self, pcs):
        """
        Add the passed PatchClampSeries to the sweep table.
        """

        kwargs = {'sweep_number': pcs.sweep_number, 'series': [pcs]}

        # FIXME appending to an existing entry would be nicer
        # but this seems to be not possible
        self.add_row(**kwargs)

    def get_series(self, sweep_number):
        """
        Return a list of PatchClampSeries for the given sweep number.
        """

        ids = self.__get_row_ids(sweep_number)

        if len(ids) == 0:
            return None

        matches = []

        for x in ids:
            for y in self[(x, 'series')]:
                matches.append(y)

        return matches

    def __get_row_ids(self, sweep_number):
        """
        Return the row ids for the given sweep number.
        """

        return [index for index, elem in enumerate(self['sweep_number'].data) if elem == sweep_number]


@register_class('IntracellularElectrodesTable', CORE_NAMESPACE)
class IntracellularElectrodesTable(DynamicTable):
    """
    Table for storing intracellular electrode related metadata'
    """
    __columns__ = (
        {'name': 'electrode',
         'description': 'Column for storing the reference to the intracellular electrode',
         'required': True,
         'index': False,
         'table': False},
    )

    @docval(*get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        # Define defaultb name and description settings
        kwargs['name'] = 'electrodes'
        kwargs['description'] = ('Table for storing intracellular electrode related metadata')
        # Initialize the DynamicTable
        super().__init__(**kwargs)


@register_class('IntracellularStimuliTable', CORE_NAMESPACE)
class IntracellularStimuliTable(DynamicTable):
    """
    Table for storing intracellular electrode related metadata'
    """
    __columns__ = (
        {'name': 'stimulus',
         'description': 'Column storing the reference to the recorded stimulus for the recording (rows)',
         'required': True,
         'index': False,
         'table': False,
         'class': TimeSeriesReferenceVectorData},
        {'name': 'stimulus_template',
         'description': 'Column storing the reference to the stimulus template for the recording (rows)',
         'required': False,
         'index': False,
         'table': False,
         'class': TimeSeriesReferenceVectorData},
    )

    @docval(*get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        # Define defaultb name and description settings
        kwargs['name'] = 'stimuli'
        kwargs['description'] = ('Table for storing intracellular stimulus related metadata')
        # Initialize the DynamicTable
        super().__init__(**kwargs)


@register_class('IntracellularResponsesTable', CORE_NAMESPACE)
class IntracellularResponsesTable(DynamicTable):
    """
    Table for storing intracellular electrode related metadata'
    """
    __columns__ = (
        {'name': 'response',
         'description': 'Column storing the reference to the recorded response for the recording (rows)',
         'required': True,
         'index': False,
         'table': False,
         'class': TimeSeriesReferenceVectorData},
    )

    @docval(*get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        # Define defaultb name and description settings
        kwargs['name'] = 'responses'
        kwargs['description'] = ('Table for storing intracellular response related metadata')
        # Initialize the DynamicTable
        super().__init__(**kwargs)


@register_class('IntracellularRecordingsTable', CORE_NAMESPACE)
class IntracellularRecordingsTable(AlignedDynamicTable):
    """
    A table to group together a stimulus and response from a single electrode and
    a single simultaneous_recording. Each row in the table represents a single recording consisting
    typically of a stimulus and a corresponding response.
    """
    @docval(*get_docval(AlignedDynamicTable.__init__, 'id', 'columns', 'colnames', 'category_tables', 'categories'))
    def __init__(self, **kwargs):
        kwargs['name'] = 'intracellular_recordings'
        kwargs['description'] = ('A table to group together a stimulus and response from a single electrode '
                                 'and a single simultaneous recording and for storing metadata about the '
                                 'intracellular recording.')
        in_category_tables = getargs('category_tables', kwargs)
        if in_category_tables is None or len(in_category_tables) == 0:
            kwargs['category_tables'] = [IntracellularElectrodesTable(),
                                         IntracellularStimuliTable(),
                                         IntracellularResponsesTable()]
            kwargs['categories'] = None
        else:
            # Check if our required data tables are supplied, otherwise add them to the list
            required_dynamic_table_given = [-1 for i in range(3)]  # The first three are our required tables
            for i, tab in enumerate(in_category_tables):
                if isinstance(tab, IntracellularElectrodesTable):
                    required_dynamic_table_given[0] = i
                elif isinstance(tab, IntracellularStimuliTable):
                    required_dynamic_table_given[1] = i
                elif isinstance(tab, IntracellularResponsesTable):
                    required_dynamic_table_given[2] = i
            # Check if the supplied tables contain data but not all required tables have been supplied
            required_dynamic_table_missing = np.any(np.array(required_dynamic_table_given[0:3]) < 0)
            if len(in_category_tables[0]) != 0 and required_dynamic_table_missing:
                raise ValueError("IntracellularElectrodeTable, IntracellularStimuliTable, and "
                                 "IntracellularResponsesTable are required when adding custom, non-empty "
                                 "tables to IntracellularRecordingsTable as the missing data for the required "
                                 "tables cannot be determined automatically")
            # Compile the complete list of tables
            dynamic_table_arg = copy(in_category_tables)
            categories_arg = [] if getargs('categories', kwargs) is None else copy(getargs('categories', kwargs))
            if required_dynamic_table_missing:
                if required_dynamic_table_given[2] < 0:
                    dynamic_table_arg.append(IntracellularResponsesTable)
                    if dynamic_table_arg[-1].name not in categories_arg:
                        categories_arg.insert(0, dynamic_table_arg[-1].name)
                if required_dynamic_table_given[1] < 0:
                    dynamic_table_arg.append(IntracellularStimuliTable())
                    if dynamic_table_arg[-1].name not in categories_arg:
                        categories_arg.insert(0, dynamic_table_arg[-1].name)
                if required_dynamic_table_given[0] < 0:
                    dynamic_table_arg.append(IntracellularElectrodesTable())
                    if dynamic_table_arg[-1].name not in categories_arg:
                        categories_arg.insert(0, dynamic_table_arg[-1].name)
            kwargs['category_tables'] = dynamic_table_arg
            kwargs['categories'] = categories_arg

        super().__init__(**kwargs)

    @docval(
        {
            "name": "electrode",
            "type": IntracellularElectrode,
            "doc": "The intracellular electrode used",
            "default": None,
        },
        {'name': 'stimulus_start_index', 'type': int, 'doc': 'Start index of the stimulus', 'default': None},
        {'name': 'stimulus_index_count', 'type': int, 'doc': 'Stop index of the stimulus', 'default': None},
        {'name': 'stimulus', 'type': TimeSeries,
         'doc': 'The TimeSeries (usually a PatchClampSeries) with the stimulus',
         'default': None},
        {'name': 'stimulus_template_start_index', 'type': int, 'doc': 'Start index of the stimulus template',
         'default': None},
        {'name': 'stimulus_template_index_count', 'type': int, 'doc': 'Stop index of the stimulus template',
         'default': None},
        {'name': 'stimulus_template', 'type': TimeSeries,
         'doc': 'The TimeSeries (usually a PatchClampSeries) with the stimulus template waveforms',
         'default': None},
        {'name': 'response_start_index', 'type': int, 'doc': 'Start index of the response', 'default': None},
        {'name': 'response_index_count', 'type': int, 'doc': 'Stop index of the response', 'default': None},
        {'name': 'response', 'type': TimeSeries,
         'doc': 'The TimeSeries (usually a PatchClampSeries) with the response',
         'default': None},
        {'name': 'electrode_metadata', 'type': dict,
         'doc': 'Additional electrode metadata to be stored in the electrodes table', 'default': None},
        {'name': 'stimulus_metadata', 'type': dict,
         'doc': 'Additional stimulus metadata to be stored in the stimuli table', 'default': None},
        {'name': 'response_metadata', 'type': dict,
         'doc': 'Additional resposnse metadata to be stored in the responses table', 'default': None},
        returns='Integer index of the row that was added to this table',
        rtype=int,
        allow_extra=True,
    )
    def add_recording(self, **kwargs):
        """
        Add a single recording to the IntracellularRecordingsTable table.

        Typically, both stimulus and response are expected. However, in some cases only a stimulus
        or a response may be recodred as part of a recording. In this case, None may be given
        for either stimulus or response, but not both. Internally, this results in both stimulus
        and response pointing to the same TimeSeries, while the start_index and index_count for
        the invalid series will both be set to -1.
        """
        # Get the input data
        stimulus_start_index, stimulus_index_count, stimulus = popargs('stimulus_start_index',
                                                                       'stimulus_index_count',
                                                                       'stimulus',
                                                                       kwargs)
        response_start_index, response_index_count, response = popargs('response_start_index',
                                                                       'response_index_count',
                                                                       'response',
                                                                       kwargs)
        electrode = popargs('electrode', kwargs)
        stimulus_template_start_index, stimulus_template_index_count, stimulus_template = popargs(
            'stimulus_template_start_index',
            'stimulus_template_index_count',
            'stimulus_template',
            kwargs)

        # if electrode is not provided, take from stimulus or response object
        if electrode is None:
            if stimulus:
                electrode = stimulus.electrode
            elif response:
                electrode = response.electrode

        # Confirm that we have at least a valid stimulus or response
        if stimulus is None and response is None:
            raise ValueError("stimulus and response cannot both be None.")

        # Compute the start and stop index if necessary
        stimulus_start_index, stimulus_index_count = self.__compute_index(stimulus_start_index,
                                                                          stimulus_index_count,
                                                                          stimulus, 'stimulus')
        response_start_index, response_index_count = self.__compute_index(response_start_index,
                                                                          response_index_count,
                                                                          response, 'response')
        stimulus_template_start_index, stimulus_template_index_count = self.__compute_index(
            stimulus_template_start_index,
            stimulus_template_index_count,
            stimulus_template, 'stimulus_template')

        # if stimulus template is already a column in the stimuli table, but stimulus_template was None
        if 'stimulus_template' in self.category_tables['stimuli'].colnames and stimulus_template is None:
            stimulus_template = stimulus if stimulus is not None else response  # set to stimulus if it was provided

        # If either stimulus or response are None, then set them to the same TimeSeries to keep the I/O happy
        response = response if response is not None else stimulus
        stimulus_provided_is_not_none = stimulus is not None  # Store if stimulus is None for error checks later
        stimulus = stimulus if stimulus_provided_is_not_none else response

        # Make sure the types are compatible.
        if ((response.neurodata_type.startswith("CurrentClamp") and
                stimulus.neurodata_type.startswith("VoltageClamp")) or
                (response.neurodata_type.startswith("VoltageClamp") and
                 stimulus.neurodata_type.startswith("CurrentClamp"))):
            raise ValueError("Incompatible types given for 'stimulus' and 'response' parameters. "
                             "'stimulus' is of type %s and 'response' is of type %s." %
                             (stimulus.neurodata_type, response.neurodata_type))
        if response.neurodata_type == 'IZeroClampSeries':
            if stimulus_provided_is_not_none:
                raise ValueError("stimulus should usually be None for IZeroClampSeries response")
        if isinstance(response, PatchClampSeries) and isinstance(stimulus, PatchClampSeries):
            # # We could also check sweep_number, but since it is mostly relevant to the deprecated SweepTable
            # # we don't really need to enforce it here
            # if response.sweep_number != stimulus.sweep_number:
            #     warnings.warn("sweep_number are usually expected to be the same for PatchClampSeries type "
            #                   "stimulus and response pairs in an intracellular recording.")
            if response.electrode != stimulus.electrode:
                raise ValueError(
                    "electrodes are usually expected to be the same for PatchClampSeries type stimulus and response "
                    "pairs in an intracellular recording."
                )

        # Compile the electrodes table data
        electrodes = copy(popargs('electrode_metadata', kwargs))
        if electrodes is None:
            electrodes = {}
        electrodes['electrode'] = electrode

        # Compile the stimuli table data
        stimuli = copy(popargs('stimulus_metadata', kwargs))
        if stimuli is None:
            stimuli = {}
        stimuli['stimulus'] = TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(
            stimulus_start_index, stimulus_index_count, stimulus)
        if stimulus_template is not None:
            stimuli['stimulus_template'] = TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(
                stimulus_template_start_index, stimulus_template_index_count, stimulus_template)

        # Compile the responses table data
        responses = copy(popargs('response_metadata', kwargs))
        if responses is None:
            responses = {}
        responses['response'] = TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(
            response_start_index, response_index_count, response)

        _ = super().add_row(enforce_unique_id=True,
                            electrodes=electrodes,
                            responses=responses,
                            stimuli=stimuli,
                            **kwargs)
        return len(self) - 1

    @staticmethod
    def __compute_index(start_index, index_count, time_series, name):
        """
        Internal helper function to compute the start_index and index_count
        to use for the stimulus and response column

        :param start_index: The start_index provided by the user
        :param index_count: The index count provided by the user
        :param time_series: The timeseries object to reference. May be None.
        :param name: Name of the table. Used only to enhance error reporting

        :raises IndexError: If index_count cannot be determined or start_index+index_count
            are outside of the range of the timeseries.

        :returns: A tuple of integers with the start_index and index_count to use.
        """
        # If times_series is not valid then return -1, -1 to indicate invalid times
        if time_series is None:
            return -1, -1
        # Since time_series is valid, negative or None start_index means the user did not specify a start_index
        # so we now need to set it to 0
        if start_index is None or start_index < 0:
            start_index = 0
        # If index_count has not been set yet (i.e., it is -1 or None) then attempt to set it to the
        # full range of the timeseries starting from start_index
        num_samples = time_series.num_samples
        if index_count is None or index_count < 0:
            index_count = (num_samples - start_index) if num_samples is not None else None
        # Check that the start_index and index_count are valid and raise IndexError if they are invalid
        if index_count is None:
            raise IndexError("Invalid %s_index_count cannot be determined from %s data." % (name, name))
        if num_samples is not None:
            if start_index >= num_samples:
                raise IndexError("%s_start_index out of range" % name)
            if (start_index + index_count) > num_samples:
                raise IndexError("%s_start_index + %s_index_count out of range" % (name, name))
        # Return the values
        return start_index, index_count

    @docval(*get_docval(AlignedDynamicTable.to_dataframe, 'ignore_category_ids'),
            {'name': 'electrode_refs_as_objectids', 'type': bool,
             'doc': 'replace object references in the electrode column with object_ids',
             'default': False},
            {'name': 'stimulus_refs_as_objectids', 'type': bool,
             'doc': 'replace object references in the stimulus column with object_ids',
             'default': False},
            {'name': 'response_refs_as_objectids', 'type': bool,
             'doc': 'replace object references in the response column with object_ids',
             'default': False}
            )
    def to_dataframe(self, **kwargs):
        """Convert the collection of tables to a single pandas DataFrame"""
        res = super().to_dataframe(ignore_category_ids=getargs('ignore_category_ids', kwargs))
        if getargs('electrode_refs_as_objectids', kwargs):
            res[('electrodes', 'electrode')] = [e.object_id for e in res[('electrodes', 'electrode')]]
        if getargs('stimulus_refs_as_objectids', kwargs):
            res[('stimuli', 'stimulus')] = \
                [e if e[2] is None
                 else TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(e[0], e[1],  e[2].object_id)
                 for e in res[('stimuli', 'stimulus')]]
        if getargs('response_refs_as_objectids', kwargs):
            res[('responses', 'response')] = \
                [e if e[2] is None else
                 TimeSeriesReferenceVectorData.TIME_SERIES_REFERENCE_TUPLE(e[0], e[1],  e[2].object_id)
                 for e in res[('responses', 'response')]]
        return res


@register_class('SimultaneousRecordingsTable', CORE_NAMESPACE)
class SimultaneousRecordingsTable(DynamicTable):
    """
    A table for grouping different intracellular recordings from the
    IntracellularRecordingsTable table together that were recorded simultaneously
    from different electrodes.
    """

    __columns__ = (
        {'name': 'recordings',
         'description': 'Column with references to one or more rows in the IntracellularRecordingsTable table',
         'required': True,
         'index': True,
         'table': True},
    )

    @docval({'name': 'intracellular_recordings_table',
             'type': IntracellularRecordingsTable,
             'doc': 'the IntracellularRecordingsTable table that the recordings column indexes. May be None when '
                    'reading the Container from file as the table attribute is already populated in this case '
                    'but otherwise this is required.',
             'default': None},
            *get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        intracellular_recordings_table = popargs('intracellular_recordings_table', kwargs)
        # Define default name and description settings
        kwargs['name'] = 'simultaneous_recordings'
        kwargs['description'] = ('A table for grouping different intracellular recordings from the'
                                 'IntracellularRecordingsTable table together that were recorded simultaneously '
                                 'from different electrodes.')
        # Initialize the DynamicTable
        super().__init__(**kwargs)
        if self['recordings'].target.table is None:
            if intracellular_recordings_table is not None:
                self['recordings'].target.table = intracellular_recordings_table
            else:
                raise ValueError("intracellular_recordings constructor argument required")

    @docval({'name': 'recordings',
             'type': 'array_data',
             'doc': 'the indices of the recordings belonging to this simultaneous recording'},
            returns='Integer index of the row that was added to this table',
            rtype=int,
            allow_extra=True)
    def add_simultaneous_recording(self, **kwargs):
        """
        Add a single simultaneous recording (i.e., one sweep, or one row) consisting of one or more
        recordings and associated custom simultaneous recording metadata to the table.
        """
        _ = super().add_row(enforce_unique_id=True, **kwargs)
        return len(self.id) - 1


@register_class('SequentialRecordingsTable', CORE_NAMESPACE)
class SequentialRecordingsTable(DynamicTable):
    """
    A table for grouping different intracellular recording simultaneous_recordings from the
    SimultaneousRecordingsTable table together. This is typically used to group together simultaneous_recordings
    where the a sequence of stimuli of the same type with varying parameters
    have been presented in a sequence.
    """

    __columns__ = (
        {'name': 'simultaneous_recordings',
         'description': 'Column with references to one or more rows in the SimultaneousRecordingsTable table',
         'required': True,
         'index': True,
         'table': True},
        {'name': 'stimulus_type',
         'description': 'Column storing the type of stimulus used for the sequential recording',
         'required': True,
         'index': False,
         'table': False}
    )

    @docval({'name': 'simultaneous_recordings_table',
             'type': SimultaneousRecordingsTable,
             'doc': 'the SimultaneousRecordingsTable table that the simultaneous_recordings '
                    'column indexes. May be None when reading the Container from file as the '
                    'table attribute is already populated in this case but otherwise this is required.',
             'default': None},
            *get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        simultaneous_recordings_table = popargs('simultaneous_recordings_table', kwargs)
        # Define defaultb name and description settings
        kwargs['name'] = 'sequential_recordings'
        kwargs['description'] = ('A table for grouping different intracellular recording simultaneous_recordings '
                                 'from the SimultaneousRecordingsTable table together. This is typically used to '
                                 'group together simultaneous_recordings where the a sequence of stimuli of the '
                                 'same type with varying parameters have been presented in a sequence.')
        # Initialize the DynamicTable
        super().__init__(**kwargs)
        if self['simultaneous_recordings'].target.table is None:
            if simultaneous_recordings_table is not None:
                self['simultaneous_recordings'].target.table = simultaneous_recordings_table
            else:
                raise ValueError('simultaneous_recordings_table constructor argument required')

    @docval({'name': 'stimulus_type',
             'type': str,
             'doc': 'the type of stimulus used for the sequential recording'},
            {'name': 'simultaneous_recordings',
             'type': 'array_data',
             'doc': 'the indices of the simultaneous_recordings belonging to this sequential recording'},
            returns='Integer index of the row that was added to this table',
            rtype=int,
            allow_extra=True)
    def add_sequential_recording(self, **kwargs):
        """
        Add a sequential recording (i.e., one row) consisting of one or more simultaneous recordings
        and associated custom sequential recording metadata to the table.
        """
        _ = super().add_row(enforce_unique_id=True, **kwargs)
        return len(self.id) - 1


@register_class('RepetitionsTable', CORE_NAMESPACE)
class RepetitionsTable(DynamicTable):
    """
    A table for grouping different intracellular recording sequential recordings together.
    With each SweepSequence typically representing a particular type of stimulus, the
    RepetitionsTable table is typically used to group sets of stimuli applied in sequence.
    """

    __columns__ = (
        {'name': 'sequential_recordings',
         'description': 'Column with references to one or more rows in the SequentialRecordingsTable table',
         'required': True,
         'index': True,
         'table': True},
    )

    @docval({'name': 'sequential_recordings_table',
             'type': SequentialRecordingsTable,
             'doc': 'the SequentialRecordingsTable table that the sequential_recordings column indexes. May '
                    'be None when reading the Container from file as the table attribute is already populated '
                    'in this case but otherwise this is required.',
             'default': None},
            *get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        sequential_recordings_table = popargs('sequential_recordings_table', kwargs)
        # Define default name and description settings
        kwargs['name'] = 'repetitions'
        kwargs['description'] = ('A table for grouping different intracellular recording sequential recordings '
                                 'together. With each SimultaneousRecording typically representing a particular type '
                                 'of stimulus, the RepetitionsTable table is typically used to group sets '
                                 'of stimuli applied in sequence.')
        # Initialize the DynamicTable
        super().__init__(**kwargs)
        if self['sequential_recordings'].target.table is None:
            if sequential_recordings_table is not None:
                self['sequential_recordings'].target.table = sequential_recordings_table
            else:
                raise ValueError('sequential_recordings_table constructor argument required')

    @docval({'name': 'sequential_recordings',
             'type': 'array_data',
             'doc': 'the indices of the sequential recordings belonging to this repetition',
             'default': None},
            returns='Integer index of the row that was added to this table',
            rtype=int,
            allow_extra=True)
    def add_repetition(self, **kwargs):
        """
        Add a repetition (i.e., one row) consisting of one or more sequential recordings
        and associated custom repetition metadata to the table.
        """
        _ = super().add_row(enforce_unique_id=True, **kwargs)
        return len(self.id) - 1


@register_class('ExperimentalConditionsTable', CORE_NAMESPACE)
class ExperimentalConditionsTable(DynamicTable):
    """
    A table for grouping different intracellular recording repetitions together that
    belong to the same experimental conditions.
    """

    __columns__ = (
        {'name': 'repetitions',
         'description': 'Column with references to one or more rows in the RepetitionsTable table',
         'required': True,
         'index': True,
         'table': True},
    )

    @docval({'name': 'repetitions_table',
             'type': RepetitionsTable,
             'doc': 'the RepetitionsTable table that the repetitions column indexes',
             'default': None},
            *get_docval(DynamicTable.__init__, 'id', 'columns', 'colnames'))
    def __init__(self, **kwargs):
        repetitions_table = popargs('repetitions_table', kwargs)
        # Define default name and description settings
        kwargs['name'] = 'experimental_conditions'
        kwargs['description'] = ('A table for grouping different intracellular recording repetitions together that '
                                 'belong to the same experimental conditions.')
        # Initialize the DynamicTable
        super().__init__(**kwargs)
        if self['repetitions'].target.table is None:
            if repetitions_table is not None:
                self['repetitions'].target.table = repetitions_table
            else:
                raise ValueError('repetitions_table constructor argument required')

    @docval({'name': 'repetitions',
             'type': 'array_data',
             'doc': 'the indices of the repetitions belonging to this condition',
             'default': None},
            returns='Integer index of the row that was added to this table',
            rtype=int,
            allow_extra=True)
    def add_experimental_condition(self, **kwargs):
        """
        Add a condition (i.e., one row) consisting of one or more repetitions of sequential recordings
        and associated custom experimental_conditions metadata to the table.
        """
        _ = super().add_row(enforce_unique_id=True, **kwargs)
        return len(self.id) - 1



# -*- coding: utf-8 -*-
"""
Calcium Imaging Data
====================

This tutorial will demonstrate how to write calcium imaging data. The workflow demonstrated here involves
five main steps:

1. Create imaging plane
2. Add acquired two-photon images
3. Add motion correction (optional)
4. Add image segmentation
5. Add fluorescence and dF/F responses

It is recommended to cover :ref:`basics` before this tutorial.

.. note:: It is recommended to check if your source data is supported by
  `NeuroConv Optical Physiology Gallery <https://neuroconv.readthedocs.io/en/main/conversion_examples_gallery/#optical-physiology>`_.
  If it is supported, it is recommended to use NeuroConv to convert your data.

The following examples will reference variables that may not be defined within the block they are used in. For
clarity, we define them here:
"""

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_ophys.png'
from datetime import datetime
from uuid import uuid4

import matplotlib.pyplot as plt
import numpy as np
from dateutil.tz import tzlocal

from pynwb import NWBHDF5IO, NWBFile, TimeSeries
from pynwb.image import ImageSeries
from pynwb.ophys import (
    CorrectedImageStack,
    Fluorescence,
    ImageSegmentation,
    MotionCorrection,
    OnePhotonSeries,
    OpticalChannel,
    RoiResponseSeries,
    TwoPhotonSeries,
)

####################
# Creating the NWB file
# ---------------------
#
# When creating an NWB file, the first step is to create the :py:class:`~pynwb.file.NWBFile` object.

nwbfile = NWBFile(
    session_description="my first synthetic recording",
    identifier=str(uuid4()),
    session_start_time=datetime.now(tzlocal()),
    experimenter=[
        "Baggins, Bilbo",
    ],
    lab="Bag End Laboratory",
    institution="University of Middle Earth at the Shire",
    experiment_description="I went on an adventure to reclaim vast treasures.",
    keywords=["ecephys", "exploration", "wanderlust"],
    related_publications="doi:10.1016/j.neuron.2016.12.011",
)

####################
# Imaging Plane
# -------------
#
# First, we must create an :py:class:`~pynwb.ophys.ImagingPlane` object, which will hold information about the area and
# method used to collect the optical imaging data. This first requires creation of a :py:class:`~pynwb.device.Device`
# object for the  microscope and an :py:class:`~pynwb.ophys.OpticalChannel` object.
#
# .. only:: html
#
#   .. image:: ../../_static/ImagingPlane.svg
#     :width: 50%
#     :alt: imaging plane UML diagram
#     :align: center
#
# .. only:: latex
#
#   .. image:: ../../_static/ImagingPlane.png
#     :width: 50%
#     :alt: imaging plane UML diagram
#     :align: center
#
# Create a :py:class:`~pynwb.device.Device` named ``"Microscope"`` in the :py:class:`~pynwb.file.NWBFile` object. Then
# create an  :py:class:`~pynwb.ophys.OpticalChannel` named ``"OpticalChannel"``.

device = nwbfile.create_device(
    name="Microscope",
    description="My two-photon microscope",
    manufacturer="The best microscope manufacturer",
)
optical_channel = OpticalChannel(
    name="OpticalChannel",
    description="an optical channel",
    emission_lambda=500.0,
)

####################
# Now, create a :py:class:`~pynwb.ophys.ImagingPlane` named ``"ImagingPlane"``, passing in the
# :py:class:`~pynwb.ophys.OpticalChannel` object and the :py:class:`~pynwb.device.Device` object.

imaging_plane = nwbfile.create_imaging_plane(
    name="ImagingPlane",
    optical_channel=optical_channel,
    imaging_rate=30.0,
    description="a very interesting part of the brain",
    device=device,
    excitation_lambda=600.0,
    indicator="GFP",
    location="V1",
    grid_spacing=[0.01, 0.01],
    grid_spacing_unit="meters",
    origin_coords=[1.0, 2.0, 3.0],
    origin_coords_unit="meters",
)

####################
# One-photon Series
# -----------------
# Now that we have our :py:class:`~pynwb.ophys.ImagingPlane`, we can create a
# :py:class:`~pynwb.ophys.OnePhotonSeries` object to store raw one-photon imaging data.

# the image data will be stored inside the NWB file
one_p_series = OnePhotonSeries(
    name="OnePhotonSeries",
    description="Raw 1p data",
    data=np.ones((1000, 100, 100)),
    imaging_plane=imaging_plane,
    rate=1.0,
    unit="normalized amplitude",
)

####################
# Since these one-photon data are acquired data, we will add the
# :py:class:`~pynwb.ophys.OnePhotonSeries` objects to the :py:class:`~pynwb.file.NWBFile`
# as acquired data.

nwbfile.add_acquisition(one_p_series)

####################
# Two-photon Series
# -----------------
# :py:class:`~pynwb.ophys.TwoPhotonSeries` objects store acquired two-photon imaging
# data. This class behaves similarly to :py:class:`~pynwb.ophys.OnePhotonSeries`.
#
# .. only:: html
#
#   .. image:: ../../_static/TwoPhotonSeries.svg
#     :width: 80%
#     :alt: two-photon series UML diagram
#     :align: center
#
# .. only:: latex
#
#   .. image:: ../../_static/TwoPhotonSeries.png
#     :width: 80%
#     :alt: two-photon series UML diagram
#     :align: center
#

# the image data will be stored inside the NWB file
two_p_series = TwoPhotonSeries(
    name="TwoPhotonSeries",
    description="Raw 2p data",
    data=np.ones((1000, 100, 100)),
    imaging_plane=imaging_plane,
    rate=1.0,
    unit="normalized amplitude",
)

nwbfile.add_acquisition(two_p_series)

####################
# Motion Correction (optional)
# ----------------------------
#
# You can also store the result of motion correction using a :py:class:`~pynwb.ophys.MotionCorrection` object,
# which is a :py:class:`~pynwb.core.MultiContainerInterface` (similar to :py:class:`~pynwb.behavior.Position`)
# which holds 1 or more :py:class:`~pynwb.ophys.CorrectedImageStack` objects.

corrected = ImageSeries(
    name="corrected",  # this must be named "corrected"
    description="A motion corrected image stack",
    data=np.ones((1000, 100, 100)),
    unit="na",
    format="raw",
    starting_time=0.0,
    rate=1.0,
)

xy_translation = TimeSeries(
    name="xy_translation",
    description="x,y translation in pixels",
    data=np.ones((1000, 2)),
    unit="pixels",
    starting_time=0.0,
    rate=1.0,
)

corrected_image_stack = CorrectedImageStack(
    corrected=corrected,
    original=one_p_series,
    xy_translation=xy_translation,
)

motion_correction = MotionCorrection(corrected_image_stacks=[corrected_image_stack])

####################
# We will create a :py:class:`~pynwb.base.ProcessingModule` named "ophys" to store optical
# physiology data and add the motion correction data to the :py:class:`~pynwb.file.NWBFile`.
#

ophys_module = nwbfile.create_processing_module(
    name="ophys", description="optical physiology processed data"
)

ophys_module.add(motion_correction)

####################
# Plane Segmentation
# ------------------
#
# The :py:class:`~pynwb.ophys.PlaneSegmentation` class stores the detected
# regions of interest in the :py:class:`~pynwb.ophys.TwoPhotonSeries` data.
# :py:class:`~pynwb.ophys.PlaneSegmentation` is a subclass of
# :py:class:`~hdmf.common.table.DynamicTable`, where each row represents a
# single region of interest (ROI).
#
# .. only:: html
#
#   .. image:: ../../_static/PlaneSegmentation.svg
#     :width: 60%
#     :alt: plane segmentation UML diagram
#     :align: center
#
# .. only:: latex
#
#   .. image:: ../../_static/PlaneSegmentation.png
#     :width: 60%
#     :alt: plane segmentation UML diagram
#     :align: center
#
# The :py:class:`~pynwb.ophys.ImageSegmentation` class can contain multiple
# :py:class:`~pynwb.ophys.PlaneSegmentation` tables, so that we can store results
# of different segmentation algorithms or different segmentation classes.
#
# .. only:: html
#
#   .. image:: ../../_static/ImageSegmentation.svg
#     :width: 100%
#     :alt: image segmentation UML diagram
#     :align: center
#
# .. only:: latex
#
#   .. image:: ../../_static/ImageSegmentation.png
#     :width: 100%
#     :alt: image segmentation UML diagram
#     :align: center
#
# First, we create an :py:class:`~pynwb.ophys.ImageSegmentation` object, then
# from that object we create a :py:class:`~pynwb.ophys.PlaneSegmentation` table
# with a link to the :py:class:`~pynwb.ophys.ImagingPlane` created earlier.
# Then we will add the :py:class:`~pynwb.ophys.ImageSegmentation` object
# to the previously created :py:class:`~pynwb.base.ProcessingModule`.

img_seg = ImageSegmentation()

ps = img_seg.create_plane_segmentation(
    name="PlaneSegmentation",
    description="output from segmenting my favorite imaging plane",
    imaging_plane=imaging_plane,
    reference_images=one_p_series,  # optional
)

ophys_module.add(img_seg)

####################
# Regions Of Interest (ROIs)
# --------------------------
#
# Image masks
# ^^^^^^^^^^^
#
# You can add ROIs to the :py:class:`~pynwb.ophys.PlaneSegmentation` table using
# an image mask or a pixel mask. An image mask is an array that is the same size
# as a single frame of the :py:class:`~pynwb.ophys.TwoPhotonSeries` that
# indicates the mask weight of each pixel in the image. Image mask values (weights) may
# be boolean or continuous between 0 and 1.


for _ in range(30):
    image_mask = np.zeros((100, 100))

    # randomly generate example image masks
    x = np.random.randint(0, 95)
    y = np.random.randint(0, 95)
    image_mask[x:x + 5, y:y + 5] = 1

    # add image mask to plane segmentation
    ps.add_roi(image_mask=image_mask)

# show one of the image masks
plt.imshow(image_mask)

####################
# Pixel masks
# ^^^^^^^^^^^
#
# Alternatively, you could define ROIs using a pixel mask, which is an array of
# triplets (x, y, weight) that have a non-zero weight. All undefined pixels are assumed
# to be 0.


ps2 = img_seg.create_plane_segmentation(
    name="PlaneSegmentation2",
    description="output from segmenting my favorite imaging plane",
    imaging_plane=imaging_plane,
    reference_images=one_p_series,  # optional
)

for _ in range(30):
    # randomly generate example starting points for region
    x = np.random.randint(0, 95)
    y = np.random.randint(0, 95)

    # define an example 4 x 3 region of pixels of weight '1'
    pixel_mask = []
    for ix in range(x, x + 4):
        for iy in range(y, y + 3):
            pixel_mask.append((ix, iy, 1))

    # add pixel mask to plane segmentation
    ps2.add_roi(pixel_mask=pixel_mask)

####################
# Voxel masks
# ^^^^^^^^^^^
#
# When storing the segmentation of volumetric imaging, you can use 3D imaging masks.
# Alternatively, you could define ROIs using a voxel mask, which is an array of
# triplets (x, y, z, weight) that have a non-zero weight. All undefined voxels are
# assumed to be 0.
#
# .. note::
#    You need to be consistent within a :py:class:`~pynwb.ophys.PlaneSegmentation` table.
#    You can add ROIs either using image masks, pixel masks, or voxel masks.

ps3 = img_seg.create_plane_segmentation(
    name="PlaneSegmentation3",
    description="output from segmenting my favorite imaging plane",
    imaging_plane=imaging_plane,
    reference_images=one_p_series,  # optional
)

from itertools import product

for _ in range(30):
    # randomly generate example starting points for region
    x = np.random.randint(0, 95)
    y = np.random.randint(0, 95)
    z = np.random.randint(0, 15)

    # define an example 4 x 3 x 2 voxel region of weight '0.5'
    voxel_mask = []
    for ix, iy, iz in product(
        range(x, x + 4),
        range(y, y + 3),
        range(z, z + 2)
    ):
        voxel_mask.append((ix, iy, iz, 0.5))

    # add voxel mask to plane segmentation
    ps3.add_roi(voxel_mask=voxel_mask)


####################
# We can view the :py:class:`~pynwb.ophys.PlaneSegmentation` table with pixel masks in
# tabular form by converting it to a :py:class:`~pandas.DataFrame`.

ps2.to_dataframe()

####################
# Storing Fluorescence Measurements
# ---------------------------------
#
# Now that the regions of interest are stored, you can store fluorescence data for these
# ROIs. This type of data is stored using the :py:class:`~pynwb.ophys.RoiResponseSeries`
# classes.
#
# .. only:: html
#
#   .. image:: ../../_static/RoiResponseSeries.svg
#     :width: 80%
#     :alt: RoiResponseSeries UML diagram
#     :align: center
#
# .. only:: latex
#
#   .. image:: ../../_static/RoiResponseSeries.png
#     :width: 80%
#     :alt: RoiResponseSeries UML diagram
#     :align: center
#
# To create a :py:class:`~pynwb.ophys.RoiResponseSeries` object, we will need to reference
# a set of rows from a :py:class:`~pynwb.ophys.PlaneSegmentation` table to
# indicate which ROIs correspond to which rows of your recorded data matrix.
# This is done using a :py:class:`~hdmf.common.table.DynamicTableRegion`, which is a type of link that
# allows you to reference specific rows of a :py:class:`~hdmf.common.table.DynamicTable`,
# such as a :py:class:`~pynwb.ophys.PlaneSegmentation` table by row indices.
#
#
# First, we create a :py:class:`~hdmf.common.table.DynamicTableRegion` that references
# the first two ROIs of the :py:class:`~pynwb.ophys.PlaneSegmentation` table.

rt_region = ps.create_roi_table_region(
    region=[0, 1], description="the first of two ROIs"
)

####################
# Then we create a :py:class:`~pynwb.ophys.RoiResponseSeries` object to store fluorescence
# data for those two ROIs.

roi_resp_series = RoiResponseSeries(
    name="RoiResponseSeries",
    description="Fluorescence responses for two ROIs",
    data=np.ones((50, 2)),  # 50 samples, 2 ROIs
    rois=rt_region,
    unit="lumens",
    rate=30.0,
)

####################
# To help data analysis and visualization tools know that this
# :py:class:`~pynwb.ophys.RoiResponseSeries` object represents fluorescence data,
# we will store the :py:class:`~pynwb.ophys.RoiResponseSeries` object inside
# of a :py:class:`~pynwb.ophys.Fluorescence` object.
# Then add the :py:class:`~pynwb.ophys.Fluorescence` object into the
# same :py:class:`~pynwb.base.ProcessingModule` named ``"ophys"`` that we created earlier.
#
# .. only:: html
#
#   .. image:: ../../_static/Fluorescence.svg
#     :width: 100%
#     :alt: fluorescence UML diagram
#     :align: center
#
# .. only:: latex
#
#   .. image:: ../../_static/Fluorescence.png
#     :width: 100%
#     :alt: fluorescence UML diagram
#     :align: center

fl = Fluorescence(roi_response_series=roi_resp_series)
ophys_module.add(fl)

####################
# .. tip::
#    If you want to store dF/F data instead of fluorescence data, then store
#    the :py:class:`~pynwb.ophys.RoiResponseSeries` object in a
#    :py:class:`~pynwb.ophys.DfOverF` object, which works the same way as the
#    :py:class:`~pynwb.ophys.Fluorescence` class.
#

####################
# Write the file
# ---------------------------------
# Once we have finished adding all of our data to our
# :py:class:`~pynwb.file.NWBFile`, make sure to write the file.
# IO operations are carried out using :py:class:`~pynwb.NWBHDF5IO`.

with NWBHDF5IO("ophys_tutorial.nwb", "w") as io:
    io.write(nwbfile)

####################
# Read the NWBFile
# ----------------
#
# We can access the raw data by indexing ``nwbfile.acquisition`` with a name
# of the :py:class:`~pynwb.ophys.TwoPhotonSeries`, e.g., ``"TwoPhotonSeries1"``.
#
# We can also access the fluorescence responses by indexing ``nwbfile.processing``
# with the name of the processing module, ``"ophys"``.
# Then, we can access the :py:class:`~pynwb.ophys.Fluorescence` object inside
# of the ``"ophys"`` processing module by indexing it with the name of the
# :py:class:`~pynwb.ophys.Fluorescence` object. The default name of
# :py:class:`~pynwb.ophys.Fluorescence` objects is ``"Fluorescence"``.
# Finally, we can access the :py:class:`~pynwb.ophys.RoiResponseSeries` object
# inside the :py:class:`~pynwb.ophys.Fluorescence` object by indexing it
# with the name of the :py:class:`~pynwb.ophys.RoiResponseSeries` object,
# which we named ``"RoiResponseSeries"``.

with NWBHDF5IO("ophys_tutorial.nwb", "r") as io:
    read_nwbfile = io.read()
    print(read_nwbfile.acquisition["TwoPhotonSeries"])
    print(read_nwbfile.processing["ophys"])
    print(read_nwbfile.processing["ophys"]["Fluorescence"])
    print(read_nwbfile.processing["ophys"]["Fluorescence"]["RoiResponseSeries"])

####################
# Accessing your data
# ------------------------------
#
# Data arrays are read passively from the file.
# Calling the data attribute on a :py:class:`~pynwb.base.TimeSeries`
# such as a :py:class:`~pynwb.ophys.RoiResponseSeries` does not read the data
# values, but presents an ``h5py`` object that can be indexed to read data.
# You can use the ``[:]`` operator to read the entire data array into memory.
# Load and print all the data values of the :py:class:`~pynwb.ophys.RoiResponseSeries`
# object representing the fluorescence data.

with NWBHDF5IO("ophys_tutorial.nwb", "r") as io:
    read_nwbfile = io.read()

    print(read_nwbfile.acquisition["TwoPhotonSeries"])
    print(read_nwbfile.processing["ophys"]["Fluorescence"]["RoiResponseSeries"].data[:])

####################
# Accessing data regions
# ------------------------------
#
# It is often preferable to read only a portion of the data. To do this, index
# or slice into the data attribute just like if you were indexing or slicing a
# :py:mod:`numpy` array.
#
# The following code prints elements ``0:10`` in the first dimension (time)
# and ``0:3`` (ROIs) in the second dimension from the fluorescence data we have written.


with NWBHDF5IO("ophys_tutorial.nwb", "r") as io:
    read_nwbfile = io.read()

    roi_resp_series = read_nwbfile.processing["ophys"]["Fluorescence"][
        "RoiResponseSeries"
    ]

    print("section of fluorescence responses:")
    print(roi_resp_series.data[0:10, 0:3])



from pynwb import NWBHDF5IO
from pynwb import validate
from pynwb.validate import get_cached_namespaces_to_validate
from pynwb.testing import TestCase
import urllib.request
import h5py
import warnings


class TestRos3Streaming(TestCase):
    """
    Test file access using the HDF5 ros3 driver.

    This test module requires h5py to be built with the ROS3 driver: conda install -c conda-forge h5py
    """
    @classmethod
    def setUpClass(cls):
        # this is the NWB Test Data dandiset #000126 sub-1/sub-1.nwb
        cls.s3_test_path = "https://dandiarchive.s3.amazonaws.com/blobs/11e/c89/11ec8933-1456-4942-922b-94e5878bb991"

    def setUp(self):
        # Skip ROS3 tests if internet is not available or the ROS3 driver is not installed
        try:
            urllib.request.urlopen('https://dandiarchive.s3.amazonaws.com/ros3test.nwb', timeout=1)
        except urllib.request.URLError:
            self.skipTest("Internet access to DANDI failed. Skipping all Ros3 streaming tests.")
        if 'ros3' not in h5py.registered_drivers():
            self.skipTest("ROS3 driver not installed. Skipping all Ros3 streaming tests.")

    def test_read(self):
        s3_path = 'https://dandiarchive.s3.amazonaws.com/ros3test.nwb'
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"Ignoring cached namespace .*",
                category=UserWarning,
            )
            with NWBHDF5IO(s3_path, mode='r', driver='ros3') as io:
                nwbfile = io.read()
                test_data = nwbfile.acquisition['ts_name'].data[:]
                self.assertEqual(len(test_data), 3)

    def test_dandi_read(self):
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore",
                message=r"Ignoring cached namespace .*",
                category=UserWarning,
            )
            with NWBHDF5IO(path=self.s3_test_path, mode='r', driver='ros3') as io:
                nwbfile = io.read()
                test_data = nwbfile.acquisition['TestData'].data[:]
                self.assertEqual(len(test_data), 3)

    def test_dandi_get_cached_namespaces(self):
        expected_namespaces = ["core"]
        expected_namespace_dependencies = {
            'core': {
                'hdmf-common': (
                    'AlignedDynamicTable',
                    'CSRMatrix',
                    'Container',
                    'Data',
                    'DynamicTable',
                    'DynamicTableRegion',
                    'ElementIdentifiers',
                    'SimpleMultiContainer',
                    'VectorData',
                    'VectorIndex'
                )
            },
            'hdmf-common': {},
            'hdmf-experimental': {
                'hdmf-common': (
                    'AlignedDynamicTable',
                    'CSRMatrix',
                    'Container',
                    'Data',
                    'DynamicTable',
                    'DynamicTableRegion',
                    'ElementIdentifiers',
                    'SimpleMultiContainer',
                    'VectorData',
                    'VectorIndex'
                )
            }
        }
        found_namespaces, _, found_namespace_dependencies = get_cached_namespaces_to_validate(
            path=self.s3_test_path, driver="ros3"
        )

        self.assertCountEqual(first=found_namespaces, second=expected_namespaces)
        self.assertDictEqual(d1=expected_namespace_dependencies, d2=expected_namespace_dependencies)

    def test_dandi_validate(self):
        result, status = validate(paths=[self.s3_test_path], driver="ros3")

        assert result == []
        assert status == 0



"""Test reading NWB files from the DANDI Archive using fsspec."""
from dandi.dandiapi import DandiAPIClient
import fsspec
import h5py
import random
import sys
import traceback

from pynwb import NWBHDF5IO


# NOTE: do not name the function with "test_" prefix, otherwise pytest
# will try to run it as a test

# TODO read dandisets systematically, not randomly
# see https://github.com/NeurodataWithoutBorders/pynwb/issues/1804

def read_first_nwb_asset():
    """Test reading the first NWB asset from a random selection of 2 dandisets that uses NWB."""
    num_dandisets_to_read = 2
    client = DandiAPIClient()
    dandisets = list(client.get_dandisets())
    random.shuffle(dandisets)
    dandisets_to_read = dandisets[:num_dandisets_to_read]
    print("Reading NWB files from the following dandisets:")
    print([d.get_raw_metadata()["identifier"] for d in dandisets_to_read])

    fs = fsspec.filesystem("http")

    failed_reads = dict()
    for i, dandiset in enumerate(dandisets_to_read):
        dandiset_metadata = dandiset.get_raw_metadata()

        # skip any dandisets that do not use NWB
        if not any(
            data_standard["identifier"] == "RRID:SCR_015242"  # this is the RRID for NWB
            for data_standard in dandiset_metadata["assetsSummary"].get("dataStandard", [])
        ):
            continue

        dandiset_identifier = dandiset_metadata["identifier"]
        print("--------------")
        print(f"{i}: {dandiset_identifier}")

        # iterate through assets until we get an NWB file (it could be MP4)
        assets = dandiset.get_assets()
        first_asset = next(assets)
        while first_asset.path.split(".")[-1] != "nwb":
            first_asset = next(assets)
        if first_asset.path.split(".")[-1] != "nwb":
            print("No NWB files?!")
            continue

        s3_url = first_asset.get_content_url(follow_redirects=1, strip_query=True)

        try:
            with fs.open(s3_url, "rb") as f:
                with h5py.File(f) as file:
                    with NWBHDF5IO(file=file) as io:
                        io.read()
        except Exception as e:
            print(traceback.format_exc())
            failed_reads[dandiset] = e

    if failed_reads:
        print(failed_reads)
        sys.exit(1)


if __name__ == "__main__":
    read_first_nwb_asset()



from hdmf.testing import TestCase, H5RoundTripMixin
from .testh5io import NWBH5IOMixin, AcquisitionH5IOMixin, NWBH5IOFlexMixin
from .utils import remove_test_file
from .icephys_testutils import create_icephys_stimulus_and_response, create_icephys_testfile

CORE_NAMESPACE = 'core'



from datetime import datetime, timedelta
from dateutil.tz import tzlocal
from collections.abc import Iterable
from warnings import warn
import copy as _copy

import numpy as np
import pandas as pd

from hdmf.common import DynamicTableRegion, DynamicTable
from hdmf.container import HERDManager
from hdmf.utils import docval, getargs, get_docval, popargs, popargs_to_dict, AllowPositional

from . import register_class, CORE_NAMESPACE
from .base import TimeSeries, ProcessingModule
from .device import Device
from .epoch import TimeIntervals
from .ecephys import ElectrodeGroup
from .icephys import (IntracellularElectrode, SweepTable, PatchClampSeries, IntracellularRecordingsTable,
                      SimultaneousRecordingsTable, SequentialRecordingsTable, RepetitionsTable,
                      ExperimentalConditionsTable)
from .image import Images
from .ophys import ImagingPlane
from .ogen import OptogeneticStimulusSite
from .misc import Units
from .core import NWBContainer, NWBDataInterface, MultiContainerInterface, ScratchData, LabelledDict


def _not_parent(arg):
    return arg['name'] != 'parent'


@register_class('LabMetaData', CORE_NAMESPACE)
class LabMetaData(NWBContainer):
    """
    Container for storing lab-specific meta-data

    The LabMetaData class serves as a base type for defining lab specific meta-data.
    To define your own lab-specific metadata, create a Neurodata Extension (NDX) for
    NWB that defines the data to add. Using the LabMetaData container as a base type
    makes it easy to add your data to an NWBFile without having to modify the NWBFile
    type itself, since adding of LabMetaData is already implemented. For more details
    on how to create an extension see the
    :nwb_overview:`Extending NWB <extensions_tutorial/extensions_tutorial_home.html>`
    tutorial.
    """

    @docval({'name': 'name', 'type': str, 'doc': 'name of lab metadata'})
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


@register_class('Subject', CORE_NAMESPACE)
class Subject(NWBContainer):
    """Subject information and metadata."""

    __nwbfields__ = (
        'age',
        "age__reference",
        'description',
        'genotype',
        'sex',
        'species',
        'subject_id',
        'weight',
        'date_of_birth',
        'strain'
    )

    @docval(
        {
            "name": "age",
            "type": (str, timedelta),
            "doc": 'The age of the subject. The ISO 8601 Duration format is recommended, e.g., "P90D" for 90 days old.'
                   'A timedelta will automatically be converted to The ISO 8601 Duration format.',
            "default": None,
        },
        {
            "name": "age__reference",
            "type": str,
            "doc": "Age is with reference to this event. Can be 'birth' or 'gestational'. If reference is omitted, "
                   "then 'birth' is implied. Value can be None when read from an NWB file with schema version "
                   "2.0 to 2.5 where age__reference is missing.",
            "default": "birth",
        },
        {
            "name": "description",
            "type": str,
            "doc": 'A description of the subject, e.g., "mouse A10".',
            "default": None,
        },
        {'name': 'genotype', 'type': str,
         'doc': 'The genotype of the subject, e.g., "Sst-IRES-Cre/wt;Ai32(RCL-ChR2(H134R)_EYFP)/wt".',
         'default': None},
        {'name': 'sex', 'type': str,
         'doc': ('The sex of the subject. Using "F" (female), "M" (male), "U" (unknown), or "O" (other) '
                 'is recommended.'), 'default': None},
        {'name': 'species', 'type': str,
         'doc': 'The species of the subject. The formal latin binomal name is recommended, e.g., "Mus musculus"',
         'default': None},
        {'name': 'subject_id', 'type': str, 'doc': 'A unique identifier for the subject, e.g., "A10"',
         'default': None},
        {'name': 'weight', 'type': (float, str),
         'doc': ('The weight of the subject, including units. Using kilograms is recommended. e.g., "0.02 kg". '
                 'If a float is provided, then the weight will be stored as "[value] kg".'),
         'default': None},
        {'name': 'date_of_birth', 'type': datetime, 'default': None,
         'doc': 'The datetime of the date of birth. May be supplied instead of age.'},
        {'name': 'strain', 'type': str, 'doc': 'The strain of the subject, e.g., "C57BL/6J"', 'default': None},
    )
    def __init__(self, **kwargs):
        keys_to_set = (
            "age",
            "age__reference",
            "description",
            "genotype",
            "sex",
            "species",
            "subject_id",
            "weight",
            "date_of_birth",
            "strain",
        )
        args_to_set = popargs_to_dict(keys_to_set, kwargs)
        super().__init__(name="subject", **kwargs)

        # NOTE when the Subject I/O mapper (see pynwb.io.file.py) reads an age__reference value of None from an
        # NWB 2.0-2.5 file, it sets the value to "unspecified" so that when Subject.__init__ is called, the incoming
        # age__reference value is NOT replaced by the default value ("birth") specified in the docval.
        # then we replace "unspecified" with None here. the user will never see the value "unspecified".
        # the ONLY way that age__reference can now be None is if it is read as None from an NWB 2.0-2.5 file.
        if self._in_construct_mode and args_to_set["age__reference"] == "unspecified":
            args_to_set["age__reference"] = None
        elif args_to_set["age__reference"] not in ("birth", "gestational"):
            raise ValueError("age__reference, if supplied, must be 'birth' or 'gestational'.")

        weight = args_to_set['weight']
        if isinstance(weight, float):
            args_to_set['weight'] = str(weight) + ' kg'

        if isinstance(args_to_set["age"], timedelta):
            args_to_set["age"] = pd.Timedelta(args_to_set["age"]).isoformat()

        date_of_birth = args_to_set['date_of_birth']
        if date_of_birth and date_of_birth.tzinfo is None:
            args_to_set['date_of_birth'] = _add_missing_timezone(date_of_birth)

        for key, val in args_to_set.items():
            setattr(self, key, val)


@register_class('NWBFile', CORE_NAMESPACE)
class NWBFile(MultiContainerInterface, HERDManager):
    """
    A representation of an NWB file.
    """

    __clsconf__ = [
        {
            'attr': 'acquisition',
            'add': '_add_acquisition_internal',
            'type': (NWBDataInterface, DynamicTable),
            'get': 'get_acquisition'
        },
        {
            'attr': 'analysis',
            'add': 'add_analysis',
            'type': (NWBContainer, DynamicTable),
            'get': 'get_analysis'
        },
        {
            'attr': 'scratch',
            'add': '_add_scratch',
            'type': (DynamicTable, NWBContainer, ScratchData),
            'get': '_get_scratch'
        },
        {
            'attr': 'stimulus',
            'add': '_add_stimulus_internal',
            'type': (NWBDataInterface, DynamicTable),
            'get': 'get_stimulus'
        },
        {
            'attr': 'stimulus_template',
            'add': '_add_stimulus_template_internal',
            'type': (TimeSeries, Images),
            'get': 'get_stimulus_template'
        },
        {
            'attr': 'processing',
            'add': 'add_processing_module',
            'type': ProcessingModule,
            'create': 'create_processing_module',
            'get': 'get_processing_module'
        },
        {
            'attr': 'devices',
            'add': 'add_device',
            'type': Device,
            'create': 'create_device',
            'get': 'get_device'
        },
        {
            'attr': 'electrode_groups',
            'add': 'add_electrode_group',
            'type': ElectrodeGroup,
            'create': 'create_electrode_group',
            'get': 'get_electrode_group'
        },
        {
            'attr': 'imaging_planes',
            'add': 'add_imaging_plane',
            'type': ImagingPlane,
            'create': 'create_imaging_plane',
            'get': 'get_imaging_plane'
        },
        {
            'attr': 'icephys_electrodes',
            'add': 'add_icephys_electrode',
            'type': IntracellularElectrode,
            'create': 'create_icephys_electrode',
            'get': 'get_icephys_electrode'
        },
        {
            'attr': 'ogen_sites',
            'add': 'add_ogen_site',
            'type': OptogeneticStimulusSite,
            'create': 'create_ogen_site',
            'get': 'get_ogen_site'
        },
        {
            'attr': 'intervals',
            'add': 'add_time_intervals',
            'type': TimeIntervals,
            'create': 'create_time_intervals',
            'get': 'get_time_intervals'
        },
        {
            'attr': 'lab_meta_data',
            'add': 'add_lab_meta_data',
            'type': LabMetaData,
            'create': 'create_lab_meta_data',
            'get': 'get_lab_meta_data'
        }
    ]

    __nwbfields__ = ({'name': 'session_description', 'settable': False},
                     {'name': 'identifier', 'settable': False},
                     {'name': 'session_start_time', 'settable': False},
                     {'name': 'timestamps_reference_time', 'settable': False},
                     {'name': 'file_create_date', 'settable': False},
                     'experimenter',
                     'experiment_description',
                     'session_id',
                     'institution',
                     'keywords',
                     'notes',
                     'pharmacology',
                     'protocol',
                     'related_publications',
                     'slices',
                     'source_script',
                     'source_script_file_name',
                     'data_collection',
                     'surgery',
                     'virus',
                     'stimulus_notes',
                     'lab',
                     {'name': 'electrodes', 'child': True,  'required_name': 'electrodes'},
                     {'name': 'epochs', 'child': True, 'required_name': 'epochs'},
                     {'name': 'trials', 'child': True, 'required_name': 'trials'},
                     {'name': 'units', 'child': True, 'required_name': 'units'},
                     {'name': 'subject', 'child': True, 'required_name': 'subject'},
                     {'name': 'sweep_table', 'child': True, 'required_name': 'sweep_table'},
                     {'name': 'invalid_times', 'child': True, 'required_name': 'invalid_times'},
                     # icephys_filtering is temporary. /intracellular_ephys/filtering dataset will be deprecated
                     {'name': 'icephys_filtering', 'settable': False},
                     {'name': 'intracellular_recordings', 'child': True,
                      'required_name': 'intracellular_recordings',
                      'doc': 'IntracellularRecordingsTable table to group together a stimulus and response '
                             'from a single intracellular electrode and a single simultaneous recording.'},
                     {'name': 'icephys_simultaneous_recordings',
                      'child': True,
                      'required_name': 'simultaneous_recordings',
                      'doc': 'SimultaneousRecordingsTable table for grouping different intracellular recordings from'
                             'the IntracellularRecordingsTable table together that were recorded simultaneously '
                             'from different electrodes'},
                     {'name': 'icephys_sequential_recordings',
                      'child': True,
                      'required_name': 'sequential_recordings',
                      'doc': 'A table for grouping different simultaneous intracellular recording from the '
                             'SimultaneousRecordingsTable table together. This is typically used to group '
                             'together simultaneous recordings where the a sequence of stimuli of the same '
                             'type with varying parameters have been presented in a sequence.'},
                     {'name': 'icephys_repetitions',
                      'child': True,
                      'required_name': 'repetitions',
                      'doc': 'A table for grouping different intracellular recording sequential recordings together.'
                             'With each SweepSequence typically representing a particular type of stimulus, the '
                             'RepetitionsTable table is typically used to group sets of stimuli applied in sequence.'},
                     {'name': 'icephys_experimental_conditions',
                      'child': True,
                      'required_name': 'experimental_conditions',
                      'doc': 'A table for grouping different intracellular recording repetitions together that '
                             'belong to the same experimental experimental_conditions.'})

    @docval({'name': 'session_description', 'type': str,
             'doc': 'a description of the session where this data was generated'},
            {'name': 'identifier', 'type': str, 'doc': 'a unique text identifier for the file'},
            {'name': 'session_start_time', 'type': datetime, 'doc': 'the start date and time of the recording session'},
            {'name': 'file_create_date', 'type': ('array_data', datetime),
             'doc': 'the date and time the file was created and subsequent modifications made', 'default': None},
            {'name': 'timestamps_reference_time', 'type': datetime,
             'doc': 'date and time corresponding to time zero of all timestamps; defaults to value '
                    'of session_start_time', 'default': None},
            {'name': 'experimenter', 'type': (tuple, list, str),
             'doc': 'name of person who performed experiment', 'default': None},
            {'name': 'experiment_description', 'type': str,
             'doc': 'general description of the experiment', 'default': None},
            {'name': 'session_id', 'type': str, 'doc': 'lab-specific ID for the session', 'default': None},
            {'name': 'institution', 'type': str,
             'doc': 'institution(s) where experiment is performed', 'default': None},
            {'name': 'keywords', 'type': 'array_data', 'doc': 'Terms to search over', 'default': None},
            {'name': 'notes', 'type': str,
             'doc': 'Notes about the experiment.', 'default': None},
            {'name': 'pharmacology', 'type': str,
             'doc': 'Description of drugs used, including how and when they were administered. '
                    'Anesthesia(s), painkiller(s), etc., plus dosage, concentration, etc.', 'default': None},
            {'name': 'protocol', 'type': str,
             'doc': 'Experimental protocol, if applicable. E.g., include IACUC protocol', 'default': None},
            {'name': 'related_publications', 'type': (tuple, list, str),
             'doc': 'Publication information.'
             'PMID, DOI, URL, etc. If multiple, concatenate together and describe which is which. '
             'such as PMID, DOI, URL, etc', 'default': None},
            {'name': 'slices', 'type': str,
             'doc': 'Description of slices, including information about preparation '
             'thickness, orientation, temperature and bath solution', 'default': None},
            {'name': 'source_script', 'type': str,
             'doc': 'Script file used to create this NWB file.', 'default': None},
            {'name': 'source_script_file_name', 'type': str,
             'doc': 'Name of the source_script file', 'default': None},
            {'name': 'data_collection', 'type': str,
             'doc': 'Notes about data collection and analysis.', 'default': None},
            {'name': 'surgery', 'type': str,
             'doc': 'Narrative description about surgery/surgeries, including date(s) '
                    'and who performed surgery.', 'default': None},
            {'name': 'virus', 'type': str,
             'doc': 'Information about virus(es) used in experiments, including virus ID, '
                    'source, date made, injection location, volume, etc.', 'default': None},
            {'name': 'stimulus_notes', 'type': str,
             'doc': 'Notes about stimuli, such as how and where presented.', 'default': None},
            {'name': 'lab', 'type': str, 'doc': 'lab where experiment was performed', 'default': None},
            {'name': 'acquisition', 'type': (list, tuple),
             'doc': 'Raw TimeSeries objects belonging to this NWBFile', 'default': None},
            {'name': 'analysis', 'type': (list, tuple),
             'doc': 'result of analysis', 'default': None},
            {'name': 'stimulus', 'type': (list, tuple),
             'doc': 'Stimulus TimeSeries, DynamicTable, or NWBDataInterface objects belonging to this NWBFile',
             'default': None},
            {'name': 'stimulus_template', 'type': (list, tuple),
             'doc': 'Stimulus template TimeSeries objects belonging to this NWBFile', 'default': None},
            {'name': 'epochs', 'type': TimeIntervals,
             'doc': 'Epoch objects belonging to this NWBFile', 'default': None},
            {'name': 'trials', 'type': TimeIntervals,
             'doc': 'A table containing trial data', 'default': None},
            {'name': 'invalid_times', 'type': TimeIntervals,
             'doc': 'A table containing times to be omitted from analysis', 'default': None},
            {'name': 'intervals', 'type': (list, tuple),
             'doc': 'any TimeIntervals tables storing time intervals', 'default': None},
            {'name': 'units', 'type': Units,
             'doc': 'A table containing unit metadata', 'default': None},
            {'name': 'processing', 'type': (list, tuple),
             'doc': 'ProcessingModule objects belonging to this NWBFile', 'default': None},
            {'name': 'lab_meta_data', 'type': (list, tuple), 'default': None,
             'doc': 'an extension that contains lab-specific meta-data'},
            {'name': 'electrodes', 'type': DynamicTable,
             'doc': 'the ElectrodeTable that belongs to this NWBFile', 'default': None},
            {'name': 'electrode_groups', 'type': Iterable,
             'doc': 'the ElectrodeGroups that belong to this NWBFile', 'default': None},
            {'name': 'ic_electrodes', 'type': (list, tuple),
             'doc': 'DEPRECATED use icephys_electrodes parameter instead. '
                    'IntracellularElectrodes that belong to this NWBFile', 'default': None},
            {'name': 'sweep_table', 'type': SweepTable,
             'doc': 'the SweepTable that belong to this NWBFile', 'default': None},
            {'name': 'imaging_planes', 'type': (list, tuple),
             'doc': 'ImagingPlanes that belong to this NWBFile', 'default': None},
            {'name': 'ogen_sites', 'type': (list, tuple),
             'doc': 'OptogeneticStimulusSites that belong to this NWBFile', 'default': None},
            {'name': 'devices', 'type': (list, tuple),
             'doc': 'Device objects belonging to this NWBFile', 'default': None},
            {'name': 'subject', 'type': Subject,
             'doc': 'subject metadata', 'default': None},
            {'name': 'scratch', 'type': (list, tuple),
             'doc': 'scratch data', 'default': None},
            {'name': 'icephys_electrodes', 'type': (list, tuple),
             'doc': 'IntracellularElectrodes that belong to this NWBFile.', 'default': None},
            {'name': 'icephys_filtering', 'type': str, 'default': None,
             'doc': '[DEPRECATED] Use IntracellularElectrode.filtering instead. Description of filtering used.'},
            {'name': 'intracellular_recordings', 'type': IntracellularRecordingsTable, 'default': None,
             'doc': 'the IntracellularRecordingsTable table that belongs to this NWBFile'},
            {'name': 'icephys_simultaneous_recordings', 'type': SimultaneousRecordingsTable, 'default': None,
             'doc': 'the SimultaneousRecordingsTable table that belongs to this NWBFile'},
            {'name': 'icephys_sequential_recordings', 'type': SequentialRecordingsTable, 'default': None,
             'doc': 'the SequentialRecordingsTable table that belongs to this NWBFile'},
            {'name': 'icephys_repetitions', 'type': RepetitionsTable, 'default': None,
             'doc': 'the RepetitionsTable table that belongs to this NWBFile'},
            {'name': 'icephys_experimental_conditions', 'type': ExperimentalConditionsTable, 'default': None,
             'doc': 'the ExperimentalConditionsTable table that belongs to this NWBFile'})
    def __init__(self, **kwargs):
        keys_to_set = [
            'session_description',
            'identifier',
            'session_start_time',
            'experimenter',
            'file_create_date',
            'ic_electrodes',
            'icephys_electrodes',
            'related_publications',
            'timestamps_reference_time',
            'acquisition',
            'analysis',
            'stimulus',
            'stimulus_template',
            'keywords',
            'processing',
            'electrodes',
            'electrode_groups',
            'devices',
            'imaging_planes',
            'ogen_sites',
            'intervals',
            'subject',
            'sweep_table',
            'lab_meta_data',
            'epochs',
            'trials',
            'invalid_times',
            'units',
            'scratch',
            'experiment_description',
            'session_id',
            'lab',
            'institution',
            'data_collection',
            'notes',
            'pharmacology',
            'protocol',
            'slices',
            'source_script',
            'source_script_file_name',
            'surgery',
            'virus',
            'stimulus_notes',
            'icephys_filtering',  # DEPRECATION warning will be raised in the setter when calling setattr in the loop
            'intracellular_recordings',
            'icephys_simultaneous_recordings',
            'icephys_sequential_recordings',
            'icephys_repetitions',
            'icephys_experimental_conditions'
        ]
        args_to_set = popargs_to_dict(keys_to_set, kwargs)
        kwargs['name'] = 'root'
        super().__init__(**kwargs)

        # add timezone to session_start_time if missing
        session_start_time = args_to_set['session_start_time']
        if session_start_time.tzinfo is None:
            args_to_set['session_start_time'] = _add_missing_timezone(session_start_time)

        # set timestamps_reference_time to session_start_time if not provided
        # if provided, ensure that it has a timezone
        timestamps_reference_time = args_to_set['timestamps_reference_time']
        if timestamps_reference_time is None:
            args_to_set['timestamps_reference_time'] = args_to_set['session_start_time']
        elif timestamps_reference_time.tzinfo is None:
            raise ValueError("'timestamps_reference_time' must be a timezone-aware datetime object.")

        # convert file_create_date to list and add timezone if missing
        file_create_date = args_to_set['file_create_date']
        if file_create_date is None:
            file_create_date = datetime.now(tzlocal())
        if isinstance(file_create_date, datetime):
            file_create_date = [file_create_date]
        args_to_set['file_create_date'] = list(map(_add_missing_timezone, file_create_date))

        # backwards-compatibility code for ic_electrodes / icephys_electrodes
        icephys_electrodes = args_to_set['icephys_electrodes']
        ic_electrodes = args_to_set['ic_electrodes']
        if icephys_electrodes is None and ic_electrodes is not None:
            warn("Use of the ic_electrodes parameter is deprecated. "
                 "Use the icephys_electrodes parameter instead", DeprecationWarning)
            args_to_set['icephys_electrodes'] = ic_electrodes
        args_to_set.pop('ic_electrodes')  # do not set this arg

        # convert single experimenter to tuple
        experimenter = args_to_set['experimenter']
        if isinstance(experimenter, str):
            args_to_set['experimenter'] = (experimenter,)

        # convert single related_publications to tuple
        related_pubs = args_to_set['related_publications']
        if isinstance(related_pubs, str):
            args_to_set['related_publications'] = (related_pubs,)

        # ensure source_script is provided if source_script_file_name is provided
        if args_to_set['source_script'] is None and args_to_set['source_script_file_name'] is not None:
            raise ValueError("'source_script' cannot be None when 'source_script_file_name' is set")

        # these attributes have no setters and can only be set using self.fields
        keys_to_set_via_fields = (
            'session_description',
            'identifier',
            'session_start_time',
            'timestamps_reference_time',
            'file_create_date'
        )
        args_to_set_via_fields = popargs_to_dict(keys_to_set_via_fields, args_to_set)

        for key, val in args_to_set_via_fields.items():
            self.fields[key] = val

        for key, val in args_to_set.items():
            setattr(self, key, val)

        self.__obj = None

    def all_children(self):
        stack = [self]
        ret = list()
        self.__obj = LabelledDict(label='all_objects', key_attr='object_id')
        while len(stack):
            n = stack.pop()
            ret.append(n)
            if n.object_id is not None:
                self.__obj[n.object_id] = n
            else:
                warn('%s "%s" does not have an object_id' % (n.neurodata_type, n.name))
            if hasattr(n, 'children'):
                for c in n.children:
                    stack.append(c)
        return ret

    @property
    def objects(self):
        if self.__obj is None:
            self.all_children()
        return self.__obj

    @property
    def modules(self):
        warn("NWBFile.modules has been replaced by NWBFile.processing.", DeprecationWarning)
        return self.processing

    @property
    def epoch_tags(self):
        return set(self.epochs.tags[:]) if self.epochs is not None else set()

    @property
    def ec_electrode_groups(self):
        warn("NWBFile.ec_electrode_groups has been replaced by NWBFile.electrode_groups.", DeprecationWarning)
        return self.electrode_groups

    @property
    def ec_electrodes(self):
        warn("NWBFile.ec_electrodes has been replaced by NWBFile.electrodes.", DeprecationWarning)
        return self.electrodes

    @property
    def ic_electrodes(self):
        warn("NWBFile.ic_electrodes has been replaced by NWBFile.icephys_electrodes.", DeprecationWarning)
        return self.icephys_electrodes

    @property
    def icephys_filtering(self):
        return self.fields.get('icephys_filtering')

    @icephys_filtering.setter
    def icephys_filtering(self, val):
        if val is not None:
            warn("Use of icephys_filtering is deprecated. Use the IntracellularElectrode.filtering field instead",
                 DeprecationWarning)
            self.fields['icephys_filtering'] = val

    def add_ic_electrode(self, *args, **kwargs):
        """
        This method is deprecated and will be removed in future versions. Please
        use :py:meth:`~pynwb.file.NWBFile.add_icephys_electrode` instead
        """
        warn("NWBFile.add_ic_electrode has been replaced by NWBFile.add_icephys_electrode.", DeprecationWarning)
        return self.add_icephys_electrode(*args, **kwargs)

    def create_ic_electrode(self, *args, **kwargs):
        """
        This method is deprecated and will be removed in future versions. Please
        use :py:meth:`~pynwb.file.NWBFile.create_icephys_electrode` instead
        """
        warn("NWBFile.create_ic_electrode has been replaced by NWBFile.create_icephys_electrode.", DeprecationWarning)
        return self.create_icephys_electrode(*args, **kwargs)

    def get_ic_electrode(self, *args, **kwargs):
        """
        This method is deprecated and will be removed in future versions. Please
        use :py:meth:`~pynwb.file.NWBFile.get_icephys_electrode` instead
        """
        warn("NWBFile.get_ic_electrode has been replaced by NWBFile.get_icephys_electrode.", DeprecationWarning)
        return self.get_icephys_electrode(*args, **kwargs)

    def __check_epochs(self):
        if self.epochs is None:
            self.epochs = TimeIntervals(name='epochs', description='experimental epochs')

    @docval(*get_docval(TimeIntervals.add_column), allow_extra=True)
    def add_epoch_column(self, **kwargs):
        """
        Add a column to the epoch table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_column` for more details
        """
        self.__check_epochs()
        self.epochs.add_column(**kwargs)

    def add_epoch_metadata_column(self, *args, **kwargs):
        """
        This method is deprecated and will be removed in future versions. Please
        use :py:meth:`~pynwb.file.NWBFile.add_epoch_column` instead
        """
        raise DeprecationWarning("Please use NWBFile.add_epoch_column")

    @docval(*get_docval(TimeIntervals.add_interval),
            allow_extra=True)
    def add_epoch(self, **kwargs):
        """

        Creates a new Epoch object. Epochs are used to track intervals
        in an experiment, such as exposure to a certain type of stimuli
        (an interval where orientation gratings are shown, or of
        sparse noise) or a different paradigm (a rat exploring an
        enclosure versus sleeping between explorations)
        """
        self.__check_epochs()
        self.epochs.add_interval(**kwargs)

    def __check_electrodes(self):
        if self.electrodes is None:
            self.electrodes = ElectrodeTable()

    @docval(*get_docval(DynamicTable.add_column), allow_extra=True)
    def add_electrode_column(self, **kwargs):
        """
        Add a column to the electrode table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_column` for more details
        """
        self.__check_electrodes()
        self.electrodes.add_column(**kwargs)

    @docval({'name': 'x', 'type': float, 'doc': 'the x coordinate of the position (+x is posterior)',
             'default': None},
            {'name': 'y', 'type': float, 'doc': 'the y coordinate of the position (+y is inferior)', 'default': None},
            {'name': 'z', 'type': float, 'doc': 'the z coordinate of the position (+z is right)', 'default': None},
            {'name': 'imp', 'type': float, 'doc': 'the impedance of the electrode, in ohms', 'default': None},
            {'name': 'location', 'type': str,
             'doc': 'the location of electrode within the subject e.g. brain region. Required.',
             'default': None},
            {'name': 'filtering', 'type': str,
             'doc': 'description of hardware filtering, including the filter name and frequency cutoffs',
             'default': None},
            {'name': 'group', 'type': ElectrodeGroup,
             'doc': 'the ElectrodeGroup object to add to this NWBFile. Required.',
             'default': None},
            {'name': 'id', 'type': int, 'doc': 'a unique identifier for the electrode', 'default': None},
            {'name': 'rel_x', 'type': float, 'doc': 'the x coordinate within the electrode group', 'default': None},
            {'name': 'rel_y', 'type': float, 'doc': 'the y coordinate within the electrode group', 'default': None},
            {'name': 'rel_z', 'type': float, 'doc': 'the z coordinate within the electrode group', 'default': None},
            {'name': 'reference', 'type': str, 'doc': 'Description of the reference electrode and/or reference scheme\
                used for this  electrode, e.g.,"stainless steel skull screw" or "online common average referencing". ',
                'default': None},
            {'name': 'enforce_unique_id', 'type': bool, 'doc': 'enforce that the id in the table must be unique',
             'default': True},
            allow_extra=True,
            allow_positional=AllowPositional.WARNING)
    def add_electrode(self, **kwargs):
        """
        Add an electrode to the electrodes table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_row` for more details.

        Required fields are *location* and
        *group* and any columns that have been added
        (through calls to `add_electrode_columns`).
        """
        self.__check_electrodes()
        d = _copy.copy(kwargs['data']) if kwargs.get('data') is not None else kwargs

        # NOTE location and group are required arguments. in PyNWB 2.1.0 we made x, y, z optional arguments, and
        # in order to avoid breaking API changes, the order of the arguments needed to be maintained even though
        # these optional arguments came before the required arguments, so in docval these required arguments are
        # displayed as optional when really they are required. this should be changed when positional arguments
        # are not allowed
        if not d['location']:
            raise ValueError("The 'location' argument is required when creating an electrode.")
        if not kwargs['group']:
            raise ValueError("The 'group' argument is required when creating an electrode.")
        if d.get('group_name', None) is None:
            d['group_name'] = d['group'].name

        new_cols = [('x', 'the x coordinate of the position (+x is posterior)'),
                    ('y', 'the y coordinate of the position (+y is inferior)'),
                    ('z', 'the z coordinate of the position (+z is right)'),
                    ('imp', 'the impedance of the electrode, in ohms'),
                    ('filtering', 'description of hardware filtering, including the filter name and frequency cutoffs'),
                    ('rel_x', 'the x coordinate within the electrode group'),
                    ('rel_y', 'the y coordinate within the electrode group'),
                    ('rel_z', 'the z coordinate within the electrode group'),
                    ('reference', 'Description of the reference electrode and/or reference scheme used for this \
                        electrode, e.g.,"stainless steel skull screw" or "online common average referencing".')
                    ]

        # add column if the arg is supplied and column does not yet exist
        # do not pass arg to add_row if arg is not supplied
        for col_name, col_doc in new_cols:
            if kwargs[col_name] is not None:
                if col_name not in self.electrodes:
                    self.electrodes.add_column(col_name, col_doc)
            else:
                d.pop(col_name)  # remove args from d if not set

        self.electrodes.add_row(**d)

    @docval({'name': 'region', 'type': (slice, list, tuple), 'doc': 'the indices of the table'},
            {'name': 'description', 'type': str, 'doc': 'a brief description of what this electrode is'},
            {'name': 'name', 'type': str, 'doc': 'the name of this container', 'default': 'electrodes'})
    def create_electrode_table_region(self, **kwargs):
        if self.electrodes is None:
            msg = "no electrodes available. add electrodes before creating a region"
            raise RuntimeError(msg)
        region = getargs('region', kwargs)
        for idx in region:
            if idx < 0 or idx >= len(self.electrodes):
                raise IndexError('The index ' + str(idx) +
                                 ' is out of range for the ElectrodeTable of length '
                                 + str(len(self.electrodes)))
        desc = getargs('description', kwargs)
        name = getargs('name', kwargs)
        return DynamicTableRegion(name=name, data=region, description=desc, table=self.electrodes)

    def __check_units(self):
        if self.units is None:
            self.units = Units(name='units', description='Autogenerated by NWBFile')

    @docval(*get_docval(Units.add_column), allow_extra=True)
    def add_unit_column(self, **kwargs):
        """
        Add a column to the unit table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_column` for more details
        """
        self.__check_units()
        self.units.add_column(**kwargs)

    @docval(*get_docval(Units.add_unit), allow_extra=True)
    def add_unit(self, **kwargs):
        """
        Add a unit to the unit table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_row` for more details.

        """
        self.__check_units()
        self.units.add_unit(**kwargs)

    def __check_trials(self):
        if self.trials is None:
            self.trials = TimeIntervals(name='trials', description='experimental trials')

    @docval(*get_docval(DynamicTable.add_column), allow_extra=True)
    def add_trial_column(self, **kwargs):
        """
        Add a column to the trial table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_column` for more details
        """
        self.__check_trials()
        self.trials.add_column(**kwargs)

    @docval(*get_docval(TimeIntervals.add_interval), allow_extra=True)
    def add_trial(self, **kwargs):
        """
        Add a trial to the trial table.
        See :py:meth:`~pynwb.epoch.TimeIntervals.add_interval` for more details.

        Required fields are *start_time*, *stop_time*, and any columns that have
        been added (through calls to `add_trial_columns`).
        """
        self.__check_trials()
        self.trials.add_interval(**kwargs)

    def __check_invalid_times(self):
        if self.invalid_times is None:
            self.invalid_times = TimeIntervals(
                name='invalid_times',
                description='time intervals to be removed from analysis'
            )

    @docval(*get_docval(DynamicTable.add_column), allow_extra=True)
    def add_invalid_times_column(self, **kwargs):
        """
        Add a column to the invalid times table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_column` for more details
        """
        self.__check_invalid_times()
        self.invalid_times.add_column(**kwargs)

    @docval(*get_docval(TimeIntervals.add_interval), allow_extra=True)
    def add_invalid_time_interval(self, **kwargs):
        """
        Add a time interval to the invalid times table.
        See :py:meth:`~hdmf.common.table.DynamicTable.add_row` for more details.

        Required fields are *start_time*, *stop_time*, and any columns that have
        been added (through calls to `add_invalid_times_columns`).
        """
        self.__check_invalid_times()
        self.invalid_times.add_interval(**kwargs)

    @docval({'name': 'electrode_table', 'type': DynamicTable, 'doc': 'the ElectrodeTable for this file'})
    def set_electrode_table(self, **kwargs):
        """
        Set the electrode table of this NWBFile to an existing ElectrodeTable
        """
        if self.electrodes is not None:
            msg = 'ElectrodeTable already exists, cannot overwrite'
            raise ValueError(msg)
        electrode_table = getargs('electrode_table', kwargs)
        self.electrodes = electrode_table

    def _check_sweep_table(self):
        """
        Create a SweepTable if not yet done.
        """
        if self.sweep_table is None:
            self.sweep_table = SweepTable(name='sweep_table')

    def _update_sweep_table(self, nwbdata):
        """
        Add all PatchClampSeries with a valid sweep number to the sweep_table
        """

        if isinstance(nwbdata, PatchClampSeries):
            if nwbdata.sweep_number is not None:
                self._check_sweep_table()
                self.sweep_table.add_entry(nwbdata)

    @docval({'name': 'nwbdata', 'type': (NWBDataInterface, DynamicTable)},
            {'name': 'use_sweep_table', 'type': bool, 'default': False, 'doc': 'Use the deprecated SweepTable'})
    def add_acquisition(self, **kwargs):
        nwbdata = popargs('nwbdata', kwargs)
        self._add_acquisition_internal(nwbdata)
        use_sweep_table = popargs('use_sweep_table', kwargs)
        if use_sweep_table:
            self._update_sweep_table(nwbdata)

    @docval({'name': 'stimulus', 'type': (TimeSeries, DynamicTable, NWBDataInterface), 'default': None,
             'doc': 'The stimulus presentation data to add to this NWBFile.'},
            {'name': 'use_sweep_table', 'type': bool, 'default': False, 'doc': 'Use the deprecated SweepTable'},
            {'name': 'timeseries', 'type': TimeSeries, 'default': None,
             'doc': 'The "timeseries" keyword argument is deprecated. Use the "nwbdata" argument instead.'},)
    def add_stimulus(self, **kwargs):
        stimulus, timeseries = popargs('stimulus', 'timeseries', kwargs)
        if stimulus is None and timeseries is None:
            raise ValueError(
                "The 'stimulus' keyword argument is required. The 'timeseries' keyword argument can be "
                "provided for backwards compatibility but is deprecated in favor of 'stimulus' and will be "
                "removed in PyNWB 3.0."
            )
        # TODO remove this support in PyNWB 3.0
        if timeseries is not None:
            warn("The 'timeseries' keyword argument is deprecated and will be removed in PyNWB 3.0. "
                 "Use the 'stimulus' argument instead.", DeprecationWarning)
            if stimulus is None:
                stimulus = timeseries
        self._add_stimulus_internal(stimulus)
        use_sweep_table = popargs('use_sweep_table', kwargs)
        if use_sweep_table:
            self._update_sweep_table(stimulus)

    @docval({'name': 'timeseries', 'type': (TimeSeries, Images)},
            {'name': 'use_sweep_table', 'type': bool, 'default': False, 'doc': 'Use the deprecated SweepTable'})
    def add_stimulus_template(self, **kwargs):
        timeseries = popargs('timeseries', kwargs)
        self._add_stimulus_template_internal(timeseries)
        use_sweep_table = popargs('use_sweep_table', kwargs)
        if use_sweep_table:
            self._update_sweep_table(timeseries)

    @docval(returns='The NWBFile.intracellular_recordings table', rtype=IntracellularRecordingsTable)
    def get_intracellular_recordings(self):
        """
        Get the NWBFile.intracellular_recordings table.

        In contrast to NWBFile.intracellular_recordings, this function will create the
        IntracellularRecordingsTable table if not yet done, whereas NWBFile.intracellular_recordings
        will return None if the table is currently not being used.
        """
        if self.intracellular_recordings is None:
            self.intracellular_recordings = IntracellularRecordingsTable()
        return self.intracellular_recordings

    @docval(*get_docval(IntracellularRecordingsTable.add_recording),
            returns='Integer index of the row that was added to IntracellularRecordingsTable',
            rtype=int,
            allow_extra=True)
    def add_intracellular_recording(self, **kwargs):
        """
        Add a intracellular recording to the intracellular_recordings table. If the
        electrode, stimulus, and/or response do not exist yet in the NWBFile, then
        they will be added to this NWBFile before adding them to the table.

        Note: For more complex organization of intracellular recordings you may also be
        interested in the related SimultaneousRecordingsTable, SequentialRecordingsTable,
        RepetitionsTable, and ExperimentalConditionsTable tables and the related functions
        of NWBFile: add_icephys_simultaneous_recording, add_icephys_sequential_recording,
        add_icephys_repetition, and add_icephys_experimental_condition.
        """
        # Add the stimulus, response, and electrode to the file if they don't exist yet
        stimulus, response, electrode = getargs('stimulus', 'response', 'electrode', kwargs)
        if (stimulus is not None and
                (stimulus.name not in self.stimulus and
                 stimulus.name not in self.stimulus_template)):
            self.add_stimulus(stimulus, use_sweep_table=False)
        if response is not None and response.name not in self.acquisition:
            self.add_acquisition(response, use_sweep_table=False)
        if electrode is not None and electrode.name not in self.icephys_electrodes:
            self.add_icephys_electrode(electrode)
        # make sure the intracellular recordings table exists and if not create it using get_intracellular_recordings
        # Add the recoding to the intracellular_recordings table
        return self.get_intracellular_recordings().add_recording(**kwargs)

    @docval(returns='The NWBFile.icephys_simultaneous_recordings table', rtype=SimultaneousRecordingsTable)
    def get_icephys_simultaneous_recordings(self):
        """
        Get the NWBFile.icephys_simultaneous_recordings table.

        In contrast to NWBFile.icephys_simultaneous_recordings, this function will create the
        SimultaneousRecordingsTable table if not yet done, whereas NWBFile.icephys_simultaneous_recordings
        will return None if the table is currently not being used.
        """
        if self.icephys_simultaneous_recordings is None:
            self.icephys_simultaneous_recordings = SimultaneousRecordingsTable(self.get_intracellular_recordings())
        return self.icephys_simultaneous_recordings

    @docval(*get_docval(SimultaneousRecordingsTable.add_simultaneous_recording),
            returns='Integer index of the row that was added to SimultaneousRecordingsTable',
            rtype=int,
            allow_extra=True)
    def add_icephys_simultaneous_recording(self, **kwargs):
        """
        Add a new simultaneous recording to the icephys_simultaneous_recordings table
        """
        return self.get_icephys_simultaneous_recordings().add_simultaneous_recording(**kwargs)

    @docval(returns='The NWBFile.icephys_sequential_recordings table', rtype=SequentialRecordingsTable)
    def get_icephys_sequential_recordings(self):
        """
        Get the NWBFile.icephys_sequential_recordings table.

        In contrast to NWBFile.icephys_sequential_recordings, this function will create the
        IntracellularRecordingsTable table if not yet done, whereas NWBFile.icephys_sequential_recordings
        will return None if the table is currently not being used.
        """
        if self.icephys_sequential_recordings is None:
            self.icephys_sequential_recordings = SequentialRecordingsTable(self.get_icephys_simultaneous_recordings())
        return self.icephys_sequential_recordings

    @docval(*get_docval(SequentialRecordingsTable.add_sequential_recording),
            returns='Integer index of the row that was added to SequentialRecordingsTable',
            rtype=int,
            allow_extra=True)
    def add_icephys_sequential_recording(self, **kwargs):
        """
        Add a new sequential recording to the icephys_sequential_recordings table
        """
        self.get_icephys_sequential_recordings()
        return self.icephys_sequential_recordings.add_sequential_recording(**kwargs)

    @docval(returns='The NWBFile.icephys_repetitions table', rtype=RepetitionsTable)
    def get_icephys_repetitions(self):
        """
        Get the NWBFile.icephys_repetitions table.

        In contrast to NWBFile.icephys_repetitions, this function will create the
        RepetitionsTable table if not yet done, whereas NWBFile.icephys_repetitions
        will return None if the table is currently not being used.
        """
        if self.icephys_repetitions is None:
            self.icephys_repetitions = RepetitionsTable(self.get_icephys_sequential_recordings())
        return self.icephys_repetitions

    @docval(*get_docval(RepetitionsTable.add_repetition),
            returns='Integer index of the row that was added to RepetitionsTable',
            rtype=int,
            allow_extra=True)
    def add_icephys_repetition(self, **kwargs):
        """
        Add a new repetition to the RepetitionsTable table
        """
        return self.get_icephys_repetitions().add_repetition(**kwargs)

    @docval(returns='The NWBFile.icephys_experimental_conditions table', rtype=ExperimentalConditionsTable)
    def get_icephys_experimental_conditions(self):
        """
        Get the NWBFile.icephys_experimental_conditions table.

        In contrast to NWBFile.icephys_experimental_conditions, this function will create the
        RepetitionsTable table if not yet done, whereas NWBFile.icephys_experimental_conditions
        will return None if the table is currently not being used.
        """
        if self.icephys_experimental_conditions is None:
            self.icephys_experimental_conditions = ExperimentalConditionsTable(self.get_icephys_repetitions())
        return self.icephys_experimental_conditions

    @docval(*get_docval(ExperimentalConditionsTable.add_experimental_condition),
            returns='Integer index of the row that was added to ExperimentalConditionsTable',
            rtype=int,
            allow_extra=True)
    def add_icephys_experimental_condition(self, **kwargs):
        """
        Add a new condition to the ExperimentalConditionsTable table
        """
        return self.get_icephys_experimental_conditions().add_experimental_condition(**kwargs)

    def get_icephys_meta_parent_table(self):
        """
        Get the top-most table in the intracellular ephys metadata table hierarchy that exists in this NWBFile.

        The intracellular ephys metadata consists of a hierarchy of DynamicTables, i.e.,
        experimental_conditions --> repetitions --> sequential_recordings -->
        simultaneous_recordings --> intracellular_recordings etc.
        In a given NWBFile not all tables may exist. This convenience functions returns the top-most
        table that exists in this file. E.g., if the file contains only the simultaneous_recordings
        and intracellular_recordings tables then the function would return the simultaneous_recordings table.
        Similarly, if the file contains all tables then it will return the experimental_conditions table.

        :returns: DynamicTable object or None
        """
        if self.icephys_experimental_conditions is not None:
            return self.icephys_experimental_conditions
        elif self.icephys_repetitions is not None:
            return self.icephys_repetitions
        elif self.icephys_sequential_recordings is not None:
            return self.icephys_sequential_recordings
        elif self.icephys_simultaneous_recordings is not None:
            return self.icephys_simultaneous_recordings
        elif self.intracellular_recordings is not None:
            return self.intracellular_recordings
        else:
            return None

    @docval({'name': 'data',
             'type': ('scalar_data', np.ndarray, list, tuple, pd.DataFrame, DynamicTable, NWBContainer, ScratchData),
             'doc': 'The data to add to the scratch space.'},
            {'name': 'name', 'type': str,
             'doc': 'The name of the data. Required only when passing in a scalar, numpy.ndarray, list, or tuple',
             'default': None},
            {'name': 'notes', 'type': str,
             'doc': ('Notes to add to the data. Only used when passing in numpy.ndarray, list, or tuple. This '
                     'argument is not recommended. Use the `description` argument instead.'),
             'default': None},
            {'name': 'table_description', 'type': str,
             'doc': ('Description for the internal DynamicTable used to store a pandas.DataFrame. This '
                     'argument is not recommended. Use the `description` argument instead.'),
             'default': ''},
            {'name': 'description', 'type': str,
             'doc': ('Description of the data. Required only when passing in a scalar, numpy.ndarray, '
                     'list, tuple, or pandas.DataFrame. Ignored when passing in an NWBContainer, '
                     'DynamicTable, or ScratchData object.'),
             'default': None})
    def add_scratch(self, **kwargs):
        '''Add data to the scratch space'''
        data, name, notes, table_description, description = getargs('data', 'name', 'notes', 'table_description',
                                                                    'description', kwargs)
        if notes is not None or table_description != '':
            warn('Use of the `notes` or `table_description` argument will be removed in a future version of PyNWB. '
                 'Use the `description` argument instead.', PendingDeprecationWarning)
            if description is not None:
                raise ValueError('Cannot call add_scratch with (notes or table_description) and description')

        if isinstance(data, (str, int, float, bytes, np.ndarray, list, tuple, pd.DataFrame)):
            if name is None:
                msg = ('A name is required for NWBFile.add_scratch when adding a scalar, numpy.ndarray, '
                       'list, tuple, or pandas.DataFrame as scratch data.')
                raise ValueError(msg)
            if isinstance(data, pd.DataFrame):
                if table_description != '':
                    description = table_description  # remove after deprecation
                if description is None:
                    msg = ('A description is required for NWBFile.add_scratch when adding a scalar, numpy.ndarray, '
                           'list, tuple, or pandas.DataFrame as scratch data.')
                    raise ValueError(msg)
                data = DynamicTable.from_dataframe(df=data, name=name, table_description=description)
            else:
                if notes is not None:
                    description = notes  # remove after deprecation
                if description is None:
                    msg = ('A description is required for NWBFile.add_scratch when adding a scalar, numpy.ndarray, '
                           'list, tuple, or pandas.DataFrame as scratch data.')
                    raise ValueError(msg)
                data = ScratchData(name=name, data=data, description=description)
        else:
            if name is not None:
                warn('The name argument is ignored when adding an NWBContainer, ScratchData, or '
                     'DynamicTable to scratch.')
            if description is not None:
                warn('The description argument is ignored when adding an NWBContainer, ScratchData, or '
                     'DynamicTable to scratch.')
        return self._add_scratch(data)

    @docval({'name': 'name', 'type': str, 'doc': 'the name of the object to get'},
            {'name': 'convert', 'type': bool, 'doc': 'return the original data, not the NWB object', 'default': True})
    def get_scratch(self, **kwargs):
        '''Get data from the scratch space'''
        name, convert = getargs('name', 'convert', kwargs)
        ret = self._get_scratch(name)
        if convert:
            if isinstance(ret, DynamicTable):
                ret = ret.to_dataframe()
            elif isinstance(ret, ScratchData):
                ret = np.asarray(ret.data)
        return ret

    def copy(self):
        """
        Shallow copy of an NWB file.
        Useful for linking across files.
        """
        kwargs = self.fields.copy()
        for key in self.fields:
            if isinstance(self.fields[key], LabelledDict):
                kwargs[key] = list(self.fields[key].values())

        # HDF5 object references cannot point to objects external to the file. Both DynamicTables such as TimeIntervals
        # contain such object references and types such as ElectricalSeries contain references to DynamicTables.
        # Below, copy the table and link to the columns so that object references work.
        fields_to_copy = ['electrodes', 'epochs', 'trials', 'units', 'sweep_table', 'invalid_times']
        for field in fields_to_copy:
            if field in kwargs:
                if isinstance(self.fields[field], DynamicTable):
                    kwargs[field] = self.fields[field].copy()
                else:
                    warn('Cannot copy child of NWBFile that is not a DynamicTable: %s' % field)

        # handle dictionaries of DynamicTables
        dt_to_copy = ['scratch', 'intervals']
        for dt in dt_to_copy:
            if dt in kwargs:
                kwargs[dt] = [v.copy() if isinstance(v, DynamicTable) else v for v in kwargs[dt]]

        return NWBFile(**kwargs)


def _add_missing_timezone(date):
    """
    Add local timezone information on a datetime object if it is missing.
    """
    if not isinstance(date, datetime):
        raise ValueError("require datetime object")
    if date.tzinfo is None:
        warn("Date is missing timezone information. Updating to local timezone.", stacklevel=2)
        return date.replace(tzinfo=tzlocal())
    return date


def _tablefunc(table_name, description, columns):
    t = DynamicTable(name=table_name, description=description)
    for c in columns:
        if isinstance(c, tuple):
            t.add_column(c[0], c[1])
        elif isinstance(c, str):
            t.add_column(c)
        else:
            raise ValueError("Elements of 'columns' must be str or tuple")
    return t


def ElectrodeTable(name='electrodes',
                   description='metadata about extracellular electrodes'):
    return _tablefunc(name, description,
                      [('location', 'the location of channel within the subject e.g. brain region'),
                       ('group', 'a reference to the ElectrodeGroup this electrode is a part of'),
                       ('group_name', 'the name of the ElectrodeGroup this electrode is a part of')
                       ]
                      )


def TrialTable(name='trials', description='metadata about experimental trials'):
    return _tablefunc(name, description, ['start_time', 'stop_time'])


def InvalidTimesTable(name='invalid_times', description='time intervals to be removed from analysis'):
    return _tablefunc(name, description, ['start_time', 'stop_time'])



"""
Defining HDF5 Dataset I/O Settings (chunking, compression, etc.)
================================================================

The HDF5 storage backend supports a broad range of advanced dataset I/O options, such as,
chunking and compression. Here we demonstrate how to use these features
from PyNWB.
"""

####################
# Wrapping data arrays with :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`
# ---------------------------------------------------------------------------
#
# In order to customize the I/O of datasets using the HDF I/O backend we simply need to wrap our datasets
# using :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`. Using H5DataIO allows us to keep the Container
# classes independent of the I/O backend while still allowing us to customize HDF5-specific I/O features.
#
# Before we get started, lets create an NWBFile for testing so that we can add our data to it.
#

from datetime import datetime

from dateutil.tz import tzlocal

from pynwb import NWBFile

start_time = datetime(2017, 4, 3, 11, tzinfo=tzlocal())

nwbfile = NWBFile(
    session_description="demonstrate advanced HDF5 I/O features",
    identifier="NWB123",
    session_start_time=start_time,
)


####################
# Normally if we create a :py:class:`~pynwb.base.TimeSeries` we would do

import numpy as np

from pynwb import TimeSeries

data = np.arange(100, 200, 10)
timestamps = np.arange(10)
test_ts = TimeSeries(
    name="test_regular_timeseries",
    data=data,
    unit="SIunit",
    timestamps=timestamps,
)
nwbfile.add_acquisition(test_ts)

####################
# Now let's say we want to chunk and compress the recorded data values. We now simply need to wrap our data with
# :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`. Everything else remains the same

from hdmf.backends.hdf5.h5_utils import H5DataIO

test_ts = TimeSeries(
    name="test_compressed_timeseries",
    data=H5DataIO(data=data, compression=True),  # <----
    unit="SIunit",
    timestamps=timestamps,
)
nwbfile.add_acquisition(test_ts)

####################
# This simple approach gives us access to a broad range of advanced I/O features, such as, chunking and
# compression. For a complete list of all available settings see :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`.
# Here, the settings for chunking and compression are determined automatically.

####################
# Chunking
# --------
#
# By default, data arrays are stored *contiguously*. This means that on disk/in memory the elements of a
# multi-dimensional, such as, ```[[1 2] [3 4]]``` are actually stored in a one-dimensional buffer
# ```1 2 3 4```. Using chunking allows us to break up our array into chunks so that our array will be
# stored not in one but multiple buffers, e.g., ``[1 2] [3 4]``. Using this approach allows optimization
# of data locality for I/O operations and enables the application of filters (e.g., compression) on a
# per-chunk basis.

#####################
# .. tip::
#
#    For an introduction to chunking and compression in HDF5 and h5py in particular see also the online book
#    `Python and HDF5 <https://www.safaribooksonline.com/library/view/python-and-hdf5/9781491944981/ch04.html>`__
#    by Andrew Collette.


####################
# To use chunking we again, simply need to wrap our dataset via :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`.
# Using chunking then also allows to also create resizable arrays simply by defining the ``maxshape`` of the array.

data = np.arange(10000).reshape((1000, 10))
wrapped_data = H5DataIO(
    data=data,
    chunks=True,  # <---- Enable chunking
    maxshape=(None, 10),  # <---- Make the time dimension unlimited and hence resizable
)
test_ts = TimeSeries(
    name="test_chunked_timeseries",
    data=wrapped_data,  # <----
    unit="SIunit",
    starting_time=0.0,
    rate=10.0,
)
nwbfile.add_acquisition(test_ts)

####################
# .. hint::
#
#   By also specifying ``fillvalue`` we can define the value that should be used when reading uninitialized
#   portions of the dataset. If no fill value has been defined, then HDF5 will use a type-appropriate default value.
#

####################
# .. note::
#
#    Chunking can help improve data read/write performance by allowing us to align chunks with common
#    read/write operations. You can find a discussion on how chunking can help in the
#    `Python and HDF5 book  <https://www.oreilly.com/library/view/python-and-hdf5/9781491944981/ch04.html>`__.
#    by Andrew Collette. But you should also know that, with great power comes great responsibility! I.e., if you choose
#    a bad chunk size e.g., too small chunks that don't align with our read/write operations, then chunking can
#    also harm I/O performance.

####################
# Compression and Other I/O Filters
# ---------------------------------
#
# HDF5 supports I/O filters, i.e, data transformation (e.g, compression) that are applied transparently on
# read/write operations.  I/O filters operate on a per-chunk basis in HDF5 and as such require the use of chunking.
# Chunking will be automatically enabled by h5py when compression and other I/O filters are enabled.
#
# To use compression, we can wrap our dataset using :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO` and
# define the appropriate options:

wrapped_data = H5DataIO(
    data=data,
    compression="gzip",  # <---- Use GZip
    compression_opts=4,  # <---- Optional GZip aggression option
)
test_ts = TimeSeries(
    name="test_gzipped_timeseries",
    data=wrapped_data,  # <----
    unit="SIunit",
    starting_time=0.0,
    rate=10.0,
)
nwbfile.add_acquisition(test_ts)

####################
# .. hint::
#
#   In addition to ``compression``, :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO` also allows us to
#   enable the ``shuffle`` and ``fletcher32`` HDF5 I/O filters.

####################
# Writing the data
# ----------------
#
#
# Writing the data now works as usual.

from pynwb import NWBHDF5IO

with NWBHDF5IO("advanced_io_example.nwb", "w") as io:
    io.write(nwbfile)

####################
# Reading the data
# ----------------
#
#
# Nothing has changed for read. All the above advanced I/O features are handled transparently.

io = NWBHDF5IO("advanced_io_example.nwb", "r")
nwbfile = io.read()

####################
# Now lets have a look to confirm that all our I/O settings where indeed used.

for k, v in nwbfile.acquisition.items():
    print(
        "name=%s, chunks=%s, compression=%s, maxshape=%s"
        % (k, v.data.chunks, v.data.compression, v.data.maxshape)
    )
io.close()
####################
#
# .. code-block:: python
#
#   name=test_regular_timeseries, chunks=None, compression=None, maxshape=(10,)
#   name=test_compressed_timeseries, chunks=(10,), compression=gzip, maxshape=(10,)
#   name=test_chunked_timeseries, chunks=(250, 5), compression=None, maxshape=(None, 10)
#   name=test_gzipped_timeseries, chunks=(250, 5), compression=gzip, maxshape=(1000, 10)
#
# As we can see, the datasets have been chunked and compressed correctly. Also, as expected, chunking
# was automatically enabled for the compressed datasets.


####################
# Wrapping ``h5py.Datasets`` with :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO`
# ---------------------------------------------------------------------------------
#
# Just for completeness, :py:class:`~hdmf.backends.hdf5.h5_utils.H5DataIO` also allows us to customize
# how ``h5py.Dataset`` objects should be handled on write by the PyNWBs HDF5 backend via the ``link_data``
# parameter. If ``link_data`` is set to ``True`` then a ``SoftLink`` or ``ExternalLink`` will be created to
# point to the HDF5 dataset On the other hand, if ``link_data`` is set to ``False`` then the dataset
# be copied using `h5py.Group.copy <https://docs.h5py.org/en/latest/high/group.html#h5py.Group.copy>`__
# **without copying attributes** and **without expanding soft links, external links, or references**.
#
# .. note::
#
#   When wrapping an ``h5py.Dataset`` object using H5DataIO, then  all settings except ``link_data``
#   will be ignored as the h5py.Dataset will either be linked to or copied as on write.

####################
# Dynamically Loaded Filters
# --------------------------
#
# HDF5 allows you to install additional filters and use these filters as plugins. Some of these filters may have
# superior performance to the default *GZIP* when it comes to read speed, write speed, or compression ratio. You can
# install several of the most popular filter plugins using the
# `hdf5plugin library <http://www.silx.org/doc/hdf5plugin/latest/index.html>`_.
#
# First, use ``pip`` to install hdf5plugin:
#
# ``pip install hdf5plugin``

####################
# This command automatically installs the filters. Here is an example of how you would use the Z Standard algorithm:

import hdf5plugin
from hdmf.backends.hdf5.h5_utils import H5DataIO

from pynwb.file import TimeSeries

wrapped_data = H5DataIO(
    data=data,
    **hdf5plugin.Zstd(clevel=3),  # set the compression and compression_opts parameters
    allow_plugin_filters=True,
)

test_ts = TimeSeries(
    name="test_gzipped_timeseries",
    data=wrapped_data,
    unit="SIunit",
    starting_time=0.0,
    rate=10.0,
)

####################
# See a list of supported compressors
# `here <http://www.silx.org/doc/hdf5plugin/latest/usage.html#write-compressed-datasets>`_.

####################
# .. note::
#
#    *h5py* (and *HDF5* more broadly) supports a number of different compression
#    algorithms, e.g., *GZIP*, *SZIP*, or *LZF* (or even custom compression filters).
#    However, only *GZIP* is built by default with HDF5, i.e., while data compressed
#    with *GZIP* can be read on all platforms and installation of HDF5, other
#    compressors may not be installed, so some users may not be able to access those
#    datasets.

####################
# Disclaimer
# ----------
#
# External links included in the tutorial are being provided as a convenience and for informational purposes only;
# they do not constitute an endorsement or an approval by the authors of any of the products, services or opinions of
# the corporation or organization or individual. The authors bear no responsibility for the accuracy, legality or
# content of the external site or for that of subsequent links. Contact the external site for answers to questions
# regarding its content.

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_h5dataio.png'



from typing import Optional

import numpy as np

from hdmf.common.table import DynamicTableRegion, DynamicTable

from ...device import Device
from ...file import ElectrodeTable, NWBFile
from ...ecephys import ElectricalSeries, ElectrodeGroup, SpikeEventSeries
from .device import mock_Device
from .utils import name_generator
from ...misc import Units


def mock_ElectrodeGroup(
    name: Optional[str] = None,
    description: str = "description",
    location: str = "location",
    device: Optional[Device] = None,
    position: Optional[str] = None,
    nwbfile: Optional[NWBFile] = None,
) -> ElectrodeGroup:

    electrode_group = ElectrodeGroup(
        name=name or name_generator("ElectrodeGroup"),
        description=description,
        location=location,
        device=device or mock_Device(nwbfile=nwbfile),
        position=position,
    )

    if nwbfile is not None:
        nwbfile.add_electrode_group(electrode_group)

    return electrode_group


def mock_ElectrodeTable(
        n_rows: int = 5, group: Optional[ElectrodeGroup] = None, nwbfile: Optional[NWBFile] = None
) -> DynamicTable:
    electrodes_table = ElectrodeTable()
    group = group if group is not None else mock_ElectrodeGroup(nwbfile=nwbfile)
    for i in range(n_rows):
        electrodes_table.add_row(
            location="CA1",
            group=group,
            group_name=group.name,
        )

    if nwbfile is not None:
        nwbfile.electrodes = electrodes_table

    return electrodes_table


def mock_electrodes(
        n_electrodes: int = 5, table: Optional[DynamicTable] = None, nwbfile: Optional[NWBFile] = None
) -> DynamicTableRegion:

    table = table or mock_ElectrodeTable(n_rows=5, nwbfile=nwbfile)
    return DynamicTableRegion(
        name="electrodes",
        data=list(range(n_electrodes)),
        description="the first and third electrodes",
        table=table,
    )


def mock_ElectricalSeries(
    name: Optional[str] = None,
    description: str = "description",
    data=None,
    rate: Optional[float] = None,
    timestamps=None,
    starting_time: Optional[float] = None,
    electrodes: Optional[DynamicTableRegion] = None,
    filtering: str = "filtering",
    nwbfile: Optional[NWBFile] = None,
    channel_conversion: Optional[np.ndarray] = None,
    conversion: float = 1.0,
    offset: float = 0.,
) -> ElectricalSeries:
    
    # Set a default rate if timestamps are not provided
    rate = 30_000.0 if (timestamps is None and rate is None) else rate

    electrical_series = ElectricalSeries(
        name=name or name_generator("ElectricalSeries"),
        description=description,
        data=data if data is not None else np.ones((10, 5)),
        rate=rate,
        starting_time=starting_time,
        timestamps=timestamps,
        electrodes=electrodes or mock_electrodes(nwbfile=nwbfile),
        filtering=filtering,
        conversion=conversion,
        offset=offset,
        channel_conversion=channel_conversion,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(electrical_series)

    return electrical_series


def mock_SpikeEventSeries(
    name: Optional[str] = None,
    description: str = "description",
    data=None,
    timestamps=np.arange(10).astype(float),
    electrodes: Optional[DynamicTableRegion] = None,
    nwbfile: Optional[NWBFile] = None,
) -> SpikeEventSeries:
    spike_event_series = SpikeEventSeries(
        name=name or name_generator("SpikeEventSeries"),
        description=description,
        data=data if data is not None else np.ones((10, 5)),
        timestamps=timestamps if timestamps is not None else np.arange(10).astype(float),
        electrodes=electrodes if electrodes is not None else mock_electrodes(nwbfile=nwbfile),
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(spike_event_series)

    return spike_event_series


def mock_Units(
    num_units: int = 10,
    max_spikes_per_unit: int = 10,
    seed: int = 0,
    nwbfile: Optional[NWBFile] = None,
) -> Units:

    units_table = Units(name="units")  # This is for nwbfile.units= mock_Units() to work
    units_table.add_column(name="unit_name", description="a readable identifier for the unit")

    rng = np.random.default_rng(seed=seed)

    times = rng.random(size=(num_units, max_spikes_per_unit)).cumsum(axis=1)
    spikes_per_unit = rng.integers(1, max_spikes_per_unit, size=num_units)

    spike_times = []
    for unit_index in range(num_units):

        # Not all units have the same number of spikes
        spike_times = times[unit_index, : spikes_per_unit[unit_index]]
        unit_name = f"unit_{unit_index}"
        units_table.add_unit(spike_times=spike_times, unit_name=unit_name)

    if nwbfile is not None:
        nwbfile.units = units_table

    return units_table



import numpy as np

from pynwb.behavior import SpatialSeries
from pynwb.testing import AcquisitionH5IOMixin, TestCase


class TestSpatialSeriesIO(AcquisitionH5IOMixin, TestCase):

    def setUpContainer(self):
        """ Return the test TimeSeries to read/write """
        return SpatialSeries(
            name='test_sS',
            data=np.ones((3, 2)),
            bounds=[(-1,1),(-1,1),(-1,1)],
            reference_frame='reference_frame',
            timestamps=[1., 2., 3.]
        )



from typing import Optional

import numpy as np

from ... import NWBFile
from ...device import Device
from ...ogen import OptogeneticStimulusSite, OptogeneticSeries

from .device import mock_Device
from .utils import name_generator


def mock_OptogeneticStimulusSite(
    name: Optional[str] = None,
    device: Optional[Device] = None,
    description: str = "optogenetic stimulus site",
    excitation_lambda: float = 500.,
    location: str = "part of the brain",
    nwbfile: Optional[NWBFile] = None,
) -> OptogeneticStimulusSite:
    optogenetic_stimulus_site = OptogeneticStimulusSite(
        name=name or name_generator("OptogeneticStimulusSite"),
        device=device or mock_Device(nwbfile=nwbfile),
        description=description,
        excitation_lambda=excitation_lambda,
        location=location
    )

    if nwbfile is not None:
        nwbfile.add_ogen_site(optogenetic_stimulus_site)

    return optogenetic_stimulus_site


def mock_OptogeneticSeries(
    name: Optional[str] = None,
    data=None,
    site: Optional[OptogeneticStimulusSite] = None,
    resolution: float = -1.0,
    conversion: float = 1.0,
    timestamps=None,
    starting_time: Optional[float] = None,
    rate: Optional[float] = 10.0,
    comments: str = "no comments",
    description: str = "no description",
    control=None,
    control_description=None,
    nwbfile: Optional[NWBFile] = None,
) -> OptogeneticSeries:
    optogenetic_series = OptogeneticSeries(
        name=name or name_generator("OptogeneticSeries"),
        data=data if data is not None else np.array([1, 2, 3, 4]),
        site=site or mock_OptogeneticStimulusSite(nwbfile=nwbfile),
        resolution=resolution,
        conversion=conversion,
        timestamps=timestamps,
        starting_time=starting_time,
        rate=rate,
        comments=comments,
        description=description,
        control=control,
        control_description=control_description,
    )

    if nwbfile is not None:
        nwbfile.add_acquisition(optogenetic_series)

    return optogenetic_series



"""
.. _scratch:

Exploratory Data Analysis with NWB
==================================

This example will focus on the basics of working with an :py:class:`~pynwb.file.NWBFile` to do more than
storing standardized data for use and exchange. For example, you may want to store results from intermediate
analyses or one-off analyses with unknown utility. This functionality is primarily accomplished with linking
and scratch space.


.. note::
    The scratch space is explicitly for non-standardized data that is not intended for reuse
    by others. Standard NWB types, and extension if required, should always be used for any data that you
    intend to share. As such, published data should not include scratch data and a user should be able
    to ignore any data stored in scratch to use a file.



"""

####################
# Raw data
# --------
#
# To demonstrate linking and scratch space, lets assume we are starting with some acquired data.
#

from datetime import datetime

import numpy as np
from dateutil.tz import tzlocal

# sphinx_gallery_thumbnail_path = 'figures/gallery_thumbnails_scratch.png'
from pynwb import NWBHDF5IO, NWBFile, TimeSeries

# set up the NWBFile
start_time = datetime(2019, 4, 3, 11, tzinfo=tzlocal())
create_date = datetime(2019, 4, 15, 12, tzinfo=tzlocal())

nwb = NWBFile(
    session_description="demonstrate NWBFile scratch",  # required
    identifier="NWB456",  # required
    session_start_time=start_time,  # required
    file_create_date=create_date,
)  # optional

# make some fake data
timestamps = np.linspace(0, 100, 1024)
data = (
    np.sin(0.333 * timestamps)
    + np.cos(0.1 * timestamps)
    + np.random.randn(len(timestamps))
)
test_ts = TimeSeries(name="raw_timeseries", data=data, unit="m", timestamps=timestamps)

# add it to the NWBFile
nwb.add_acquisition(test_ts)

with NWBHDF5IO("raw_data.nwb", mode="w") as io:
    io.write(nwb)

####################
# .. _basic_copying:
#
# Copying an NWB file
# -------------------
#
# To copy a file, we must first read the file.
#
raw_io = NWBHDF5IO("raw_data.nwb", "r")
nwb_in = raw_io.read()

####################
# And then create a shallow copy the file with the :py:func:`~pynwb.file.NWBFile.copy` method
# of :py:class:`~pynwb.file.NWBFile` .

nwb_proc = nwb_in.copy()

####################
#
# Now that we have a copy, lets process some data, and add the results as a :py:class:`~pynwb.base.ProcessingModule`
# to our copy of the file. [#]_

import scipy.signal as sps

mod = nwb_proc.create_processing_module(
    "filtering_module", "a module to store filtering results"
)

ts1 = nwb_in.acquisition["raw_timeseries"]
filt_data = sps.correlate(ts1.data, np.ones(128), mode="same") / 128
ts2 = TimeSeries(name="filtered_timeseries", data=filt_data, unit="m", timestamps=ts1)

mod.add_container(ts2)


####################
#
# Now write the copy, which contains the processed data. [#]_

with NWBHDF5IO("processed_data.nwb", mode="w", manager=raw_io.manager) as io:
    io.write(nwb_proc)


####################
#
# .. [#]
#    .. note::
#       Notice here that we are reusing the timestamps to the original TimeSeries.
#
# .. [#]
#    .. note::
#       The ``processed_data.nwb`` file (i.e., our copy) stores our processing module and contains external
#       links to all data in our original file, i.e., the data from our raw file is being linked to, not copied.
#       This allows us to isolate our processing data in a separate file while still allowing us to access the
#       raw data from our ``processed_data.nwb`` file, without having to duplicate the data.
#


####################
# .. _basic_scratch:
#
# Adding scratch data
# -------------------
#
# You may end up wanting to store results from some one-off analysis, and writing an extension
# to get your data into an NWBFile is too much over head. This is facilitated by the scratch space
# in NWB. [#]_
#
# First, lets read our processed data and then make a copy

proc_io = NWBHDF5IO("processed_data.nwb", "r")
nwb_proc_in = proc_io.read()

####################
#
# Now make a copy to put our scratch data into [#]_

nwb_scratch = nwb_proc_in.copy()

####################
#
# Now lets do an analysis for which we do not have a specification, but we would like to store
# the results for.

filt_ts = nwb_scratch.modules["filtering_module"]["filtered_timeseries"]

fft = np.fft.fft(filt_ts.data)

nwb_scratch.add_scratch(
    fft,
    name="dft_filtered",
    description="discrete Fourier transform from filtered data",
)


####################
#
# Finally, write the results.

with NWBHDF5IO("scratch_analysis.nwb", "w", manager=proc_io.manager) as io:
    io.write(nwb_scratch)

####################
#
# To get your results back, you can index into :py:attr:`~pynwb.file.NWBFile.scratch` or use
# :py:func:`~pynwb.file.NWBFile.get_scratch`:

scratch_io = NWBHDF5IO("scratch_analysis.nwb", "r")
nwb_scratch_in = scratch_io.read()

fft_in = nwb_scratch_in.scratch["dft_filtered"]

fft_in = nwb_scratch_in.get_scratch("dft_filtered")

####################
#
# .. [#]
#    .. note::
#       This scratch space only exists if you add scratch data.
#
# .. [#]
#    .. note::
#       We recommend writing scratch data into copies of files only. This will make it easier to
#       isolate and discard scratch data and avoids updating files that store precious data.

# close the IO objects
raw_io.close()
proc_io.close()
scratch_io.close()



"""Tests related to pynwb.io.utils."""
import pytest

from hdmf.build import GroupBuilder
from pynwb.io.utils import get_nwb_version
from pynwb.testing import TestCase


class TestGetNWBVersion(TestCase):

    def test_get_nwb_version(self):
        """Get the NWB version from a builder."""
        builder1 = GroupBuilder(name="root")
        builder1.set_attribute(name="nwb_version", value="2.0.0")
        builder2 = GroupBuilder(name="another")
        builder1.set_group(builder2)
        assert get_nwb_version(builder1) == (2, 0, 0)
        assert get_nwb_version(builder2) == (2, 0, 0)

    def test_get_nwb_version_missing(self):
        """Get the NWB version from a builder where the root builder does not have an nwb_version attribute."""
        builder1 = GroupBuilder(name="root")
        builder2 = GroupBuilder(name="another")
        builder1.set_group(builder2)

        with pytest.raises(ValueError, match="'nwb_version' attribute is missing from the root of the NWB file."):
            get_nwb_version(builder1)

        with pytest.raises(ValueError, match="'nwb_version' attribute is missing from the root of the NWB file."):
            get_nwb_version(builder1)

    def test_get_nwb_version_prerelease_false(self):
        """Get the NWB version from a builder."""
        builder1 = GroupBuilder(name="root")
        builder1.set_attribute(name="nwb_version", value="2.0.0-alpha")
        assert get_nwb_version(builder1) == (2, 0, 0)

    def test_get_nwb_version_prerelease_true1(self):
        """Get the NWB version from a builder."""
        builder1 = GroupBuilder(name="root")
        builder1.set_attribute(name="nwb_version", value="2.0.0-alpha")
        assert get_nwb_version(builder1, include_prerelease=True) == (2, 0, 0, "alpha")

    def test_get_nwb_version_prerelease_true2(self):
        """Get the NWB version from a builder."""
        builder1 = GroupBuilder(name="root")
        builder1.set_attribute(name="nwb_version", value="2.0.0-alpha.sha-test.5114f85")
        assert get_nwb_version(builder1, include_prerelease=True) == (2, 0, 0, "alpha.sha-test.5114f85")

    def test_get_nwb_version_20b(self):
        """Get the NWB version from a builder where version == "2.0b"."""
        builder1 = GroupBuilder(name="root")
        builder1.set_attribute(name="nwb_version", value="2.0b")
        assert get_nwb_version(builder1) == (2, 0, 0)
        assert get_nwb_version(builder1, include_prerelease=True) == (2, 0, 0, "b")
